//Copyright 2017-2018 blacknoise.at
//
//   Licensed under the Apache License, Version 2.0 (the "License");
//   you may not use this file except in compliance with the License.
//   You may obtain a copy of the License at
//
//       http://www.apache.org/licenses/LICENSE-2.0
//
//   Unless required by applicable law or agreed to in writing, software
//   distributed under the License is distributed on an "AS IS" BASIS,
//   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//   See the License for the specific language governing permissions and
//   limitations under the License.
//
#include <stdio.h>
//#include <stdlib.h>
//#include <string.h>
//#include <strings.h>
#include <locale.h>

#include <bncsql.h>
//#define bncSQL_DEBUG
#include <albus_log.h>
#ifndef bncSQL_DEBUG
# undef  ER
# define ER
# undef  MI
# define MI
# undef  D8
# define D8
# undef  D9
# define D9
#endif

long  mssql_API_Version = bncSQL_API_VERSION;
char  mssql_Version[]   = "1.0.0";

//#undef _THREAD_SAFE
//#undef TDS_HAVE_PTHREAD_MUTEX

#define _THREAD_SAFE 1
#define TDS_HAVE_PTHREAD_MUTEX 1

//#define TDS_HAVE_MUTEX 1
#include <freetds/tds.h>
#include <freetds/convert.h>


extern "C" {

static char       *Error   = NULL; //2do overlay main.cpp Error
static TDSSOCKET  *tds     = NULL; //2do mv 2 local
       TDSCONTEXT *context = NULL; //2do mv 2 local


static int tsql_handle_message ( const TDSCONTEXT * context, TDSSOCKET * tds, TDSMESSAGE * msg ) {
    D9 ( "Error=%p msg->msgno=%d msg->message=~%s~", Error, msg->msgno, msg->message );
    if ( ! Error ) { return 0; }

    if ( msg->msgno == 5703 ) return 0;//[05703]Changed language setting to us_english.
    if ( msg->msgno == 5701 ) return 0;//[05701]Der Datenbankkontext wurde auf 'master' geaendert.
    //2do if ( msg->msgno == 2403 ) return 0;//char convert ...

    //TDS_CHAR *server;
    //TDS_CHAR *message;
    //TDS_CHAR *proc_name;
    //TDS_CHAR *sql_state;
    //TDS_UINT msgno;
    //TDS_INT line_number;
    //TDS_SMALLINT state;
    //TDS_TINYINT priv_msg_type;
    //TDS_TINYINT severity;
    //int oserr;
    sprintf ( Error, "[%05d]%s\n", msg->msgno, msg->message );
    //2dofprintf(stderr, "Msg %d, Level %d, State %d, Server %s, Line %d\n%s\n", msg->msgno, msg->severity, msg->state, msg->server, msg->line_number, msg->message );

    if ( msg->msgno == 0 ) return 0;

    return TDS_INT_CANCEL;

}//static int tsql_handle_message(const TDSCONTEXT * context, TDSSOCKET * tds, TDSMESSAGE * msg)


char *mssql_dbOpen ( _bncDataBase *db ) {
    D9 ( "db=%p", db );

    Error = db->Error;
    if ( Error[0] ) return Error;

//mssql defaults
    db->Quote	= ' ';			//use ~ ~ in SQL

    if ( db->Conn ) return Error;

    if ( ! db->Port ) db->Port = 1433;

    D2 ( "db->Host=%s" , db->Host );
    D2 ( "db->Port=%ld", db->Port );
    D2 ( "db->Base=%s" , db->Base );
    D2 ( "db->User=%s" , db->User );
    D9 ( "db->Pass=%s" , db->Pass );//D9

//login 2 database
TDSLOGIN  *login = tds_alloc_login (0);
    if ( ! login ) SER ( "! login" );

    tds_set_server		( login, db->Host       );
    tds_set_port                ( login, db->Port       );
    //                                   db->Base done via USE
    tds_set_user		( login, db->User	);
    tds_set_passwd		( login, db->Pass	);
    //tds_set_passwd		( login, "db->Pass"	);//bug4debug, 2do hang @ proxy
    //tds_set_client_charset	( login, "ISO-8859-1"	);
    tds_set_client_charset	( login, "UTF-8"	);
    //tds_set_language		( login, "us_english"	);
    //tds_set_library		( login, "TDS-Library"	);

    context		 = tds_alloc_context ( NULL );
    context->msg_handler = tsql_handle_message;
    context->err_handler = tsql_handle_message;

    free ( context->locale->date_fmt );
           //ntext->locale->date_fmt = strdup ( "%d.%m.%Y %H:%M:%S.%z"	);//2do
           context->locale->date_fmt = strdup ( "%Y-%m-%d %H:%M:%S.%z"	);

    tds = tds_alloc_socket ( context, 512 );
    tds_set_parent ( tds, NULL );

TDSLOGIN *connection = tds_read_config_info ( NULL, login, context->locale );
    if ( ! connection ) SER ( "! connection" );

    tds_connect_and_login ( tds, connection ); SIER ( Error );
    tds_free_login ( connection );
    tds_free_login ( login      );

char          sqlcmd[255];
    sprintf ( sqlcmd, "USE %s", db->Base );
    tds_submit_query   ( tds, sqlcmd ); SIER ( Error );
int result_type;
    tds_process_tokens ( tds, &result_type, NULL, TDS_TOKEN_RESULTS );
    SIER ( Error );

    db->Conn = &tds;
//all OK

    return NULL;

}//char *mssql_dbOpen ( _bncDataBase *db )


char *mssql_dbClose ( _bncDataBase *db ) {
    D9 ( "db=%p", db );

    Error = db->Error;

//close aktiv conn
    if ( ! db->Conn ) return Error;

    D9 ( " db->Base=%s", db->Base );
    D9 ( " db->Host=%s", db->Host );

    tds_free_socket  ( tds     );
                       tds = NULL;
    tds_free_context ( context );
                       context = NULL;
    db->Conn = NULL;

    return NULL;

}//char *mssql_dbClose ( _bncDataBase *db )


char *mssql_tbOpen ( _bncTable *tb ) {
    D9 ( "tb=%p", tb );

    Error = tb->Error;
    if ( Error[0] ) return NULL;

//error check
    if ( ! tb->Query    ) SER ( "! tb->Query"    );
    if ( ! tb->Query[0] ) SER ( "! tb->Query[0]" );
    if (   tb->Result   ) return NULL;

    D9 ( "tb->Query=~%s~", tb->Query );

int ret = 0;
int result_type = 0;

    ret = tds_submit_query   ( tds, tb->Query ); if ( ret != TDS_SUCCESS ) SER ( "TDS_SUCCEED != %ld", ret );
    SIER ( Error );

    ret = tds_process_tokens ( tds, &result_type, NULL, TDS_TOKEN_RESULTS ); if ( ret != TDS_SUCCESS ) SER ( "TDS_SUCCEED != %ld", ret );
    SIER ( Error );
    if ( result_type != 4049 ) return NULL;	//! SELECT
    if ( result_type != TDS_ROWFMT_RESULT ) SER ( "expected row fmt()" );

    ret = tds_process_tokens ( tds, &result_type, NULL, TDS_TOKEN_RESULTS ); if ( ret != TDS_SUCCESS ) SER ( "TDS_SUCCEED != %ld", ret );
    SIER ( Error );
    if ( result_type != TDS_ROW_RESULT ) return NULL;

//fetch in first record
    ret = tds_process_tokens ( tds, &result_type, NULL, TDS_RETURN_ROW ); if ( ret != TDS_SUCCESS ) SER ( "TDS_SUCCEED != %ld", ret );
    SIER ( Error );
    if ( ret != TDS_NO_MORE_RESULTS && ret != TDS_SUCCESS) SER ( "tds_process_tokens() unexpected return" );

//2do reconnect if error
    tb->Result = (void *)&tb->Records;//2do, use cursor

//error check
    if ( Error[0] ) return NULL;

//amount of Records
    tb->Records = -1;
    tb->Fokus   =  0;
//amount of Fields
    tb->Fields  = tds->current_results->num_cols;

    return NULL;

}//char *mssql_tbOpen ( _bncTable *tb )


_bncColumn *mssql_tbColumnInsert ( _bncTable *tb, _bncColumn *co, unsigned long ColumnInsertFokus ) {
    D9 ( "tb=%p", tb );

    Error = tb->Error;
    if ( Error[0] ) return NULL;

TDSCOLUMN *curcol = tds->current_results->columns[ColumnInsertFokus];

    if ( ( SYBINTN == curcol->column_type ) && ( 8 == curcol->column_size ) ) curcol->column_type = SYBINT8;
    D9 ( "column_name=~%20s~ column_type=%05d column_size=%05d column_prec=%03d, column_scale=%03d", curcol->column_name->dstr_s, curcol->column_type, curcol->column_size, curcol->column_prec, curcol->column_scale );


//long myFlags = 0;2do

    co->Name = strdup ( curcol->column_name->dstr_s );

//Flags Mapping
    //2do if ( strcmp ( myName, "ID" ) == 0 ) myFlags += bnCO_PRI_KEY;
    //2do if ( strcmp ( myName, "ID" ) == 0 ) myFlags += bnCO_AUTO_INCREMENT;

    if ( 1 == 0 ) {}
//TDS_SERVER_TYPE;
    else if ( SYBCHAR               == curcol->column_type ) { co->Type |= bn_char; *co->Size = curcol->column_size;                                     }//SYBCHAR               =  47, 0x2F
    else if ( SYBVARCHAR            == curcol->column_type ) { co->Type |= bn_char; *co->Size = curcol->column_size;                                     }//SYBVARCHAR            =  39, 0x27
    else if ( SYBINTN               == curcol->column_type ) { co->Type |= bn_long;                                                                      }//SYBINTN               =  38, 0x26
    //2do	SYBINT1               =  48, 0x30
    else if ( SYBINT2               == curcol->column_type ) { co->Type |= bn_long;                                                                      }//SYBINT2               =  52, 0x34 2do create
    else if ( SYBINT4               == curcol->column_type ) { co->Type |= bn_long;                                                                      }//SYBINT4               =  56, 0x38 2do create
    //2do	SYBFLT8               =  62, 0x3E
    else if ( SYBDATETIME           == curcol->column_type ) { co->Type |= bn_char;   *co->Size = 23;                                                    }//SYBDATETIME           =  61, 0x3D 2do create
    //2do	SYBBIT                =  50, 0x32
    else if ( SYBTEXT               == curcol->column_type ) { co->Type |= bn_blob;                                                                      }//SYBTEXT               =  35, 0x23
    //2do	SYBNTEXT              =  99, 0x63 
    else if ( SYBIMAGE              == curcol->column_type ) { co->Type |= bn_blob;                                                                      }//SYBIMAGE              =  34, 0x22
    //2do	SYBMONEY4             = 122, 0x7A
    //2do	SYBMONEY              =  60, 0x3C
    //2do	SYBDATETIME4          =  58, 0x3A
    //2do	SYBREAL               =  59, 0x3B
    else if ( SYBBINARY             == curcol->column_type ) { co->Type |= bn_blob;                                                                      }//SYBBINARY             =  45, 0x2D
    //2do	SYBVOID               =  31, 0x1F
    else if ( SYBVARBINARY          == curcol->column_type ) { co->Type |= bn_blob;                                                                      }//SYBVARBINARY          =  37, 0x25
    else if ( SYBBITN               == curcol->column_type ) { co->Type |= bn_long;                                                                      }//SYBBITN               = 104, 0x68
    else if ( SYBNUMERIC            == curcol->column_type ) { co->Type |= bn_double; *co->Size = curcol->column_prec; co->Scale = curcol->column_scale; }//SYBNUMERIC            = 108, 0x6C
    else if ( SYBDECIMAL            == curcol->column_type ) { co->Type |= bn_double; *co->Size = curcol->column_prec; co->Scale = curcol->column_scale; }//SYBDECIMAL            = 106, 0x6A
    else if ( SYBFLTN               == curcol->column_type ) { co->Type |= bn_double; *co->Size =                  17; co->Scale =                   17; }//SYBFLTN               = 109, 0x6D 2do @ bncSQLTable 17,17 if set -> show = static
    else if ( SYBMONEYN             == curcol->column_type ) { co->Type |= bn_double; *co->Size = curcol->column_prec; co->Scale =                    4; }//SYBMONEYN             = 110, 0x6E 2do = 4, read from locale
    else if ( SYBDATETIMN           == curcol->column_type ) { co->Type |= bn_char;   *co->Size = 23;                                                    }//SYBDATETIMN           = 111, 0x6F
//MS only types
    //2do	SYBNVARCHAR           = 103, 0x67
    else if ( SYBINT8               == curcol->column_type ) { co->Type |= bn_long_long;                                                                 }//SYBINT8               = 127, 0x7F
    //2do	XSYBCHAR              = 175, 0xAF
    //2do	XSYBVARCHAR           = 167, 0xA7
    //2do	XSYBNVARCHAR          = 231, 0xE7
    //2do	XSYBNCHAR             = 239, 0xEF
    //2do	XSYBVARBINARY         = 165, 0xA5
    //2do	XSYBBINARY            = 173, 0xAD
    else if ( SYBUNIQUE             == curcol->column_type ) { co->Type |= bn_char;   *co->Size = 36;                                                    }//SYBUNIQUE             =  36, 0x24
    //2do	SYBVARIANT            =  98, 0x62
    //2do	SYBMSUDT              = 240, 0xF0
    //2do	SYBMSXML              = 241, 0xF1
    else if ( SYBMSDATE             == curcol->column_type ) { co->Type |= bn_char;   *co->Size = 19;                                                    }//SYBMSDATE             =  40, 0x28
    else if ( SYBMSTIME             == curcol->column_type ) { co->Type |= bn_char;   *co->Size = 27;                                                    }//SYBMSTIME             =  41, 0x29
    //2do	SYBMSDATETIME2        =  42, 0x2a
    //2do	SYBMSDATETIMEOFFSET   =  43, 0x2b
//Sybase only types
    //2do	SYBLONGBINARY         = 225, 0xE1
    //2do	SYBUINT1              =  64, 0x40
    //2do	SYBUINT2              =  65, 0x41
    //2do	SYBUINT4              =  66, 0x42
    //2do	SYBUINT8              =  67, 0x43
    //2do	SYBBLOB               =  36, 0x24
    //2do	SYBBOUNDARY           = 104, 0x68
    //2do	SYBDATE               =  49, 0x31
    //2do	SYBDATEN              = 123, 0x7B
    //2do	SYB5INT8              = 191, 0xBF
    //2do	SYBINTERVAL           =  46, 0x2E
    //2do	SYBLONGCHAR           = 175, 0xAF
    //2do	SYBSENSITIVITY        = 103, 0x67
    //2do	SYBSINT1              = 176, 0xB0
    //2do	SYBTIME               =  51, 0x33
    //2do	SYBTIMEN              = 147, 0x93
    //2do	SYBUINTN              =  68, 0x44
    //2do	SYBUNITEXT            = 174, 0xAE
    //2do	SYBXML                = 163, 0xA3
    //2do	SYB5BIGDATETIME       = 187, 0xBB
    //2do	SYB5BIGTIME           = 188, 0xBC
    else SER ( "unknown type=%03d name=~%s~", curcol->column_type, curcol->column_name->dstr_s );

    D9 ( " co->Type=0x%x", co->Type );
    return co;

}//_bncColumn *mssql_tbColumnInsert ( _bncTable *tb, _bncColumn *co, unsigned long ColumnInsertFokus )


char *mssql_tbClose ( _bncTable *tb ) {
    D9 ( "tb=%p", tb );

    Error = tb->Error;
    if ( ! tb->Result ) return NULL;

//2do free
    tb->Result = NULL;

    return NULL;

}//char *mssql_tbClose ( _bncTable *tb )


char *mssql_tbColumn ( _bncTable *tb, const char *ColumnName, unsigned char **ret_data, unsigned long *ret_len ) {
    D9 ( "tb=%p", tb );

    Error = tb->Error;
    if ( Error[0]        ) return NULL;
    if ( ! tb->Records   ) return NULL;
    if ( ! ColumnName    ) SER ( "! ColumnName"    );
    if ( ! ColumnName[0] ) SER ( "! ColumnName[0]" );
    if ( ! ret_data      ) SER ( "! ret_data"      );
    if ( ! ret_len       ) SER ( "! ret_len"       );

    setlocale ( LC_ALL, "C" );

    D9 ( "ColumnName=%s", ColumnName );
    D9 ( "ret_data  =%p", ret_data   );
    D9 ( "ret_len   =%p", ret_len    );

//show tables work around
char         tb_FetchColumn_Name[255];
    strcpy ( tb_FetchColumn_Name, ColumnName );
    if ( strcmp ( tb->Query, "show tables" ) == 0 )
	if ( strcmp ( tb_FetchColumn_Name, "name" ) == 0 )
	    sprintf ( tb_FetchColumn_Name, "Tables_in_%s", tb->db->Base );
//show tables work around done

long xxcnt=0;
TDSCOLUMN *curcol = NULL;
    while ( curcol = tds->current_results->columns[xxcnt] ) {
	D9 ( "column_type=%05d column_size=%05d column_name=~%s~", curcol->column_type, curcol->column_size, curcol->column_name->dstr_s );
	if ( strcmp ( curcol->column_name->dstr_s, tb_FetchColumn_Name ) == 0 ) break;
	xxcnt++;
    }//while ( curcol = tds->current_results->columns[xxcnt] )
    if ( ! curcol ) return NULL;
    D9 ( "column_type=%05d column_size=%05d column_name=~%s~", curcol->column_type, curcol->column_size, curcol->column_name->dstr_s );
    D9 ( "curcol->column_cur_size=%05d", curcol->column_cur_size );

//data -> NULL
    if ( curcol->column_cur_size < 0 ) return NULL;

TDS_CHAR *src = (TDS_CHAR *)curcol->column_data;
int conv_type = tds_get_conversion_type ( curcol->column_type, curcol->column_size );

    if ( is_blob_type ( curcol->column_type ) ) {//SYBTEXT, SYBIMAGE, SYBNTEXT
    TDSBLOB *blob = (TDSBLOB *)src;
	src = blob->textvalue;
	*ret_data = (unsigned char *)src;
	*ret_len  = curcol->column_cur_size;
	return NULL;
    }

    if (
	   ( SYBVARBINARY == curcol->column_type )
	|| ( SYBBINARY    == curcol->column_type )
    ) {
	*ret_data = (unsigned char *)src;
	*ret_len  = curcol->column_cur_size;
	return NULL;
    }


CONV_RESULT cr;
    if ( tds_convert( context, conv_type, src, curcol->column_cur_size, SYBVARCHAR, &cr ) < 0 ) return NULL;
    D9 ( "cr.c=~%s~", cr.c );

//float 2do bn_float -> mv 2 bncSQLTable...
//2do    if ( SYBFLTN == curcol->column_type ) {
//2do    _bncColumn *co = (*tb)[ColumnName];
//2do	if ( co ) {
//2do	    D9 ( "co->Name=%s", co->Name );
//2do	    *co->Size  =  strlen ( cr.c );
//2do	     co->Scale =                0;
//2do	    for ( unsigned int crPos = 0; crPos < strlen (cr.c); crPos++ ) {
//2do		if ( '.' == cr.c[crPos] ) {
//2do		    D9 ( "crPos=%ld", crPos );
//2do		    *co->Size  = crPos;
//2do		     co->Scale = strlen ( cr.c ) - *co->Size - 1;
//2do		    D7 ( "*co->Size=%ld co->Scale=%ld", *co->Size, co->Scale );//2do @ lib...
//2do		    break;
//2do		}
//2do	    }//for ( unsigned int crPos = 0; crPos < strlen (cr.c); crPos++ )
//2do	}//if ( co )
//2do    }//if ( SYBFLTN == curcol->column_type )

    *ret_data = (unsigned char *)cr.c; //new bncsql
    *ret_len  =                     0; //new bncsql

    return NULL;

}//char *mssql_tbColumn ( _bncTable *tb, const char *ColumnName, unsigned char **ret_data, unsigned long *ret_len )


char *mssql_tbRecord ( _bncTable *tb ) {
    D9 ( "tb=%p", tb );

    Error = tb->Error;
    if ( Error[0] ) return NULL;

    if ( tb->Fokus == 1 ) return NULL;//2do ?!

int result_type;
    tds_process_tokens ( tds, &result_type, NULL, TDS_RETURN_ROW );
    SIER ( Error );
    if ( result_type == 4052 ) {
	tb->Records = 0;
	return Error;
    }

    return NULL;

}//char *mssql_tbRecord ( _bncTable *tb )


char *mssql_tbSQLBuild ( _bncTable *tb ) {
    D9 ( "tb=%p", tb );

    Error = tb->Error;
    if ( Error[0] ) return NULL;

    if ( strncmp ( tb->Query, "show tables", strlen ( "show tables" ) ) == 0 ) {
	bn_free  ( tb->Query );
	           tb->Query = strdup ( "select name as name from sys.tables order by name;" );
    }

    return NULL;

}//char *mssql_tbSQLBuild ( _bncTable *tb )


char *mssql_coDEF ( _bncColumn *co ) {
    D9 ( "co=%p", co );

    Error = co->Error;
    SER ( "src it out" ); //bug2
    if ( Error[0] ) return NULL;

static char  myDef[512];
    memset ( myDef, 0x00, 512 );
    if ( 1 == 0 ) {}
    //2do MySQL    if      ( bn_long      & co->Type ) { sprintf ( myDef , "%s %10s int4"        , myDef, co->Name            ); }
    //2do MySQL    else if ( bn_long_long & co->Type ) { sprintf ( myDef , "%s %10s bigint"      , myDef, co->Name            ); }
    //2do MySQL    else if ( bn_double    & co->Type ) { sprintf ( myDef , "%s %10s float8"      , myDef, co->Name            ); }
    //2do MySQL    else if ( bn_blob      & co->Type ) { sprintf ( myDef , "%s %10s blob"        , myDef, co->Name            ); }
    //2do MySQL    else if ( bn_char      & co->Type ) { sprintf ( myDef , "%s %10s varchar(%ld)", myDef, co->Name, *co->Size ); }
    //2do    "serial"	,   bnLONG, 0, "serial PRIMARY KEY", 0, 0 , from mssql
    //2do    "double"	, bnDOUBLE, 0,             "float8", 0, 0 , from mssql
    //2do    "char"	,   bnCHAR, 0,       "varchar(%ld)", 0, 0 ,         from mssql
    else SER ( "unknown co->Type=0x%08x co->Name=~%s~", co->Type, co->Name );
    //2do MySQL    if ( co->Type & bn_PRI_KEY        ) sprintf ( myDef , "%s PRIMARY KEY"   , myDef );
    //2do MySQL    if ( co->Type & bn_AUTO_INCREMENT ) sprintf ( myDef , "%s auto_increment", myDef );
    return myDef;

}//char *mssql_coDEF ( _bncColumn *co )


}//extern "C"
