//Copyright 2016-2018 blacknoise.at
//
//   Licensed under the Apache License, Version 2.0 (the "License");
//   you may not use this file except in compliance with the License.
//   You may obtain a copy of the License at
//
//       http://www.apache.org/licenses/LICENSE-2.0
//
//   Unless required by applicable law or agreed to in writing, software
//   distributed under the License is distributed on an "AS IS" BASIS,
//   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//   See the License for the specific language governing permissions and
//   limitations under the License.
//
//#define DEBUG
//#define bnForward_DEBUG
//#define HAVE_SSL
#ifdef HAVE_CONFIG
# include <config.h>
#endif
#include "bnForward.h"
#include <stdio.h>

#include <string>
#include <sstream>
#include <vector>

using namespace std;

#define StatusServerOpen        0x0002
#define StatusClientOpen        0x0004
#define StatusConfigRequest     0x0008
#define StatusConfig            0x0010
#define StatusConnectRequest    0x0020
#define StatusConnect           0x0040
#define StatusConnectedRequest  0x0080
#define StatusConnected         0x0100

//#define DEBUG
#include <albus_log.h>
#ifndef bnForward_DEBUG
# undef  MI
# define MI
# undef  D8
# define D8
# undef  D9
# define D9
#endif


long soServer_SaveLoop = 0;
long soRemote_SaveLoop = 0;


static vector<string> fwSplit( const string &sSplit, char delim ) {//2do via bnSplitPos
vector<string> items;
string item;
stringstream ss ( sSplit );
    while ( getline ( ss, item, delim ) ) items.push_back ( item );

    return items;
}


static void fwReplace ( std::string &str, const std::string &strOut, const std::string& str_In ) {//2do via ???
    if ( ! strOut.length() ) return;
int pos = 0;
    while ( ( pos = str.find ( strOut, pos ) ) != (int)std::string::npos ) {
	str.replace ( pos, strOut.length(), str_In );
	pos += str_In.length ();
    }
}


void *bnForward__AcceptCB ( _bnSocket *cbSocket, void *cbArg ) {
    D9 ( "cbSocket=%p cbArg=%p", cbSocket, cbArg );

_bnForwardMeta *Meta = new _bnForwardMeta;
    cbSocket->Meta = Meta;

    Meta->LocalRX  = new _bnBlob;
    Meta->LocalTX  = new _bnBlob;
    Meta->RemoteRX = new _bnBlob;
    Meta->RemoteTX = new _bnBlob;

char *Error = Meta->Error;
_bnForward     *bnForward = (_bnForward *)cbArg;
    Meta->soLocal  =      cbSocket;
    Meta->soRemote = new _bnSocket;
    D2 ( "bnForward=%p Meta->soRemote=%p new", bnForward, Meta->soRemote );
    if ( bnForward->RemoteHost              ) { Meta->soRemote->Host              = strdup ( bnForward->RemoteHost              ); }
                                                Meta->soRemote->Port              =          bnForward->RemotePort               ;
                                                Meta->soRemote->TLS               =          bnForward->RemoteTLS                ;
    if ( bnForward->RemoteTLSCertificate    ) { Meta->soRemote->TLSCertificate    = strdup ( bnForward->RemoteTLSCertificate    ); }
    if ( bnForward->RemoteTLSPrivateKey     ) { Meta->soRemote->TLSPrivateKey     = strdup ( bnForward->RemoteTLSPrivateKey     ); }
    if ( bnForward->RemoteTLSPrivateKeyPass ) { Meta->soRemote->TLSPrivateKeyPass = strdup ( bnForward->RemoteTLSPrivateKeyPass ); }
    if ( bnForward->RemoteTLSFingerprint00  ) { Meta->soRemote->TLSFingerprint00  = strdup ( bnForward->RemoteTLSFingerprint00  ); }
    if ( bnForward->RemoteTLSFingerprint01  ) { Meta->soRemote->TLSFingerprint01  = strdup ( bnForward->RemoteTLSFingerprint01  ); }
    if ( bnForward->RemoteTLSFingerprint02  ) { Meta->soRemote->TLSFingerprint02  = strdup ( bnForward->RemoteTLSFingerprint02  ); }
    if ( bnForward->RemoteTLSFingerprint03  ) { Meta->soRemote->TLSFingerprint03  = strdup ( bnForward->RemoteTLSFingerprint03  ); }
    if ( bnForward->RemoteTLSFingerprint04  ) { Meta->soRemote->TLSFingerprint04  = strdup ( bnForward->RemoteTLSFingerprint04  ); }
                                                Meta->soRemote->rx                = Meta->RemoteRX;
                                                Meta->soRemote->tx                = Meta->RemoteTX;
	//	                                Meta->soRemote->SaveFile = strdup ( "/tmp/bnForward-soClient" );//bug4debug
	//	                                Meta->soRemote->SaveLoop = &soServer_SaveLoop;//bug4debug
	//	                                Meta->soRemote->SaveLoop = &soRemote_SaveLoop;//bug4debug

    Meta->Status |= StatusServerOpen;

time_t now =  0;
    time ( &now );
    Meta->Timeout = now + 10;//10 sec timeout
    Meta->soLocal->rx = Meta->LocalRX;
    Meta->soLocal->tx = Meta->LocalTX;
    Meta->soRemote->Open ();
    if ( Meta->soRemote->Error[0] ) {
	Meta->soLocal->fdClose = 1;
	SER ( "Meta->soRemote=%p Meta->soRemote->Error=%s", Meta->soRemote, Meta->soRemote->Error );
    }
    if ( Meta->soRemote->fdOpen ) {
	Meta->Status |= StatusClientOpen;
    }
    else {
	Meta->soLocal->fdClose = 1;
	SER ( "Meta->soRemote=%p Meta->soRemote->Error=%s", Meta->soRemote, Meta->soRemote->Error );
    }

    if ( ! bnForward->ProxyConnection ) {
	D7 ( "Meta=%p ! ProxyConnection", Meta );
	Meta->Status |= StatusConfigRequest   ;
	Meta->Status |= StatusConfig          ;
	Meta->Status |= StatusConnectRequest  ;
	Meta->Status |= StatusConnect         ;
	Meta->Status |= StatusConnectedRequest;
    }//if ( ! bnForward->ProxyConnection )

    return NULL;

}//void *bnForward__AcceptCB ( _bnSocket *cbSocket, void *cbArg )


void *bnForward__CloseCB ( _bnSocket *cbSocket, void *cbArg ) {
    D9 ( "cbSocket=%p cbSocket->fdSocket=%03ld cbArg=%p", cbSocket, cbSocket->fdSocket, cbArg );

    if ( ! cbSocket->Meta ) return NULL;

_bnForwardMeta *Meta = (_bnForwardMeta *)cbSocket->Meta;

    if ( Meta->soRemote ) {
	delete Meta->soRemote;
	       Meta->soRemote = NULL;
    }

    delete Meta;
    cbSocket->Meta = NULL;

_bnForward     *bnForward = (_bnForward *)cbArg;
    if ( bnForward->CloseAndExit ) {
	D7 ( "Meta=%p CloseAndExit", Meta );//D7
	strcpy ( bnForward->Error, "CloseAndExit" );
	D9 ( "Meta=%p %s", Meta, bnForward->Error );
    }//if ( bnForward->CloseAndExit )

    MI ( "close" );
    return NULL;

}//void *bnForward__CloseCB ( _bnSocket *cbSocket, void *cbArg )


char *_bnForward::Set ( long *Idx, const std::string &Key, long *KeyIdx, const std::string &Val ) {
    D9 ( "Idx=~%p~ Key=~%s~ KeyIdx=~%p~ Val=~%s~", Idx, Key.c_str(), KeyIdx, Val.c_str() );

    if ( 1 == 0 ) {}
    else if ( "RemoteHost"             == Key ) {
	if ( strncmp ( Val.c_str(), "https://", strlen ("https://") ) == 0 ) {
	    RemoteHost = strdup ( Val.c_str() + strlen ("https://") );
	    RemoteTLS  =   1;
	    RemotePort = 443;
	}
	else if ( strncmp ( Val.c_str(), "http://", strlen ("http://") ) == 0 ) {
	    RemoteHost = strdup ( Val.c_str() + strlen ("http://") );
	    RemotePort =  80;
	}
	else {
	    RemoteHost = strdup ( Val.c_str() );
	}
    }
    else if ( "RemotePort"              == Key ) RemotePort              = atol   ( Val.c_str() );
    else if ( "RemoteTLS"               == Key ) RemoteTLS               = atol   ( Val.c_str() );
    else if ( "RemoteTLSCertificate"    == Key ) RemoteTLSCertificate    = strdup ( Val.c_str() );
    else if ( "RemoteTLSPrivateKey"     == Key ) RemoteTLSPrivateKey     = strdup ( Val.c_str() );
    else if ( "RemoteTLSPrivateKeyPass" == Key ) RemoteTLSPrivateKeyPass = strdup ( Val.c_str() );
    else if ( "RemoteTLSFingerprint00"  == Key ) RemoteTLSFingerprint00  = strdup ( Val.c_str() );
    else if ( "RemoteTLSFingerprint01"  == Key ) RemoteTLSFingerprint01  = strdup ( Val.c_str() );
    else if ( "RemoteTLSFingerprint02"  == Key ) RemoteTLSFingerprint02  = strdup ( Val.c_str() );
    else if ( "RemoteTLSFingerprint03"  == Key ) RemoteTLSFingerprint03  = strdup ( Val.c_str() );
    else if ( "RemoteTLSFingerprint04"  == Key ) RemoteTLSFingerprint04  = strdup ( Val.c_str() );
    else if ( "LocalListen"             == Key ) LocalListen             = atol   ( Val.c_str() );
    else if ( "LocalHost"               == Key ) LocalHost               = strdup ( Val.c_str() );
    else if ( "LocalPort"               == Key ) LocalPort               = atol   ( Val.c_str() );
    else if ( "LocalTLS"                == Key ) LocalTLS                = atol   ( Val.c_str() );
    else if ( "LocalTLSCertificate"     == Key ) LocalTLSCertificate     = strdup ( Val.c_str() );
    else if ( "LocalTLSPrivateKey"      == Key ) LocalTLSPrivateKey      = strdup ( Val.c_str() );
    else if ( "LocalTLSPrivateKeyPass"  == Key ) LocalTLSPrivateKeyPass  = strdup ( Val.c_str() );
    else if ( "ProxyConnection"         == Key ) ProxyConnection         = atol   ( Val.c_str() );
    else if ( "UserAgent"               == Key ) UserAgent               = strdup ( Val.c_str() );
    else if ( "ConnectRequest"          == Key ) ConnectRequest          = strdup ( Val.c_str() );
    else if ( "PrintPID"                == Key ) PrintPID                = atol   ( Val.c_str() );
    else if ( "PrintPort"               == Key ) PrintPort               = atol   ( Val.c_str() );
    else if ( "CloseAndExit"            == Key ) CloseAndExit            = atol   ( Val.c_str() );

    return NULL;

}//char *_bnForward::Set ( long *Idx, const std::string &Key, long *KeyIdx, const std::string &Val )



char *_bnForward::Loop ( void ) {
    D9 ( "this=%p", this );

    if ( ! InitDone ) {

	if ( ! UserAgent     ) UserAgent     = strdup ( "bnForward/1.0.0" );
	if ( ! ConnectRequest ) ConnectRequest = strdup ( "bnforward"       );

	                                (*soServer).Listen            =          LocalListen             ;//Server
	if ( LocalHost              ) { (*soServer).Host              = strdup ( LocalHost              ); }
	                                (*soServer).Port              =          LocalPort               ;
	                                (*soServer).TLS               =          LocalTLS                ;
	if ( LocalTLSCertificate    ) { (*soServer).TLSCertificate    = strdup ( LocalTLSCertificate    ); }
	if ( LocalTLSPrivateKey     ) { (*soServer).TLSPrivateKey     = strdup ( LocalTLSPrivateKey     ); }
	if ( LocalTLSPrivateKeyPass ) { (*soServer).TLSPrivateKeyPass = strdup ( LocalTLSPrivateKeyPass ); }
	//	                        (*soServer).SaveFile = strdup ( "/tmp/bnForward-soServer" );//bug4debug
	//	                        (*soServer).SaveLoop = &soServer_SaveLoop;//bug4debug
	(*soServer).Error[0] = 0;
	(*soServer).Open ();
	(*soServer).AcceptCB    = &bnForward__AcceptCB;
	(*soServer).AcceptCBArg = this;
	(*soServer).CloseCB     = &bnForward__CloseCB;
	(*soServer).CloseCBArg  = this;

	if ( PrintPort ) {
	    fprintf ( stdout, "Port=%d\n", (*soServer).Port );
	    fflush  ( stdout );
	}
	if ( PrintPID ) {
	    fprintf ( stdout, "PID=%d\n", getpid () );
	    fflush  ( stdout );
	}

	if ( ! LocalPort ) LocalPort = (*soServer).Port;
	D8 ( "RemoteHost    =~%20s~", RemoteHost     );
	D8 ( "RemotePort    = %5ld" , RemotePort     );
	D8 ( "RemoteTLS     = %5ld" , RemoteTLS      );
	D8 ( "LocalHost     =~%20s~", LocalHost      );
	D8 ( "LocalPort     = %5ld" , LocalPort      );
	D8 ( "UserAgent     =~%s~"  , UserAgent      );
	D8 ( "ConnectRequest=~%s~"  , ConnectRequest );
	D8 ( "PrintPID      = %5ld" , PrintPID       );
	D8 ( "PrintPort     = %5ld" , PrintPort      );
	D8 ( "CloseAndExit  = %5ld" , CloseAndExit   );
	MI ( "Init" );
	InitDone = 1;
	SIER ( (*soServer).Error );
    }//if ( ! InitDone )

time_t now =  0;
    time ( &now );

    for ( unsigned int vs_stack_pos = 0; vs_stack_pos < (*(*soServer).vs_stack).size(); vs_stack_pos++ ) {
	D9 ( "this=%p vs_stack_pos=%03ld", this, vs_stack_pos );

    _bnSocket *fdServer = (*(*soServer).vs_stack)[vs_stack_pos];
    _bnForwardMeta *Meta = (_bnForwardMeta *)fdServer->Meta;
    char *Error = Meta->Error;
	D9 ( "Meta=%p", Meta );
	if ( ! Meta ) continue; // 2do PEC ( "! fdServer->Meta" );

    _bnSocket   *fdClient  = Meta->soRemote;
	if ( ! Meta->Status ) continue;

	if ( ( Meta->Timeout ) && ( now > Meta->Timeout ) ) {
	    Meta->soLocal->fdClose = 1;
	    PEC ( "Meta=%p timeout", Meta );
	}

	D9 ( "Meta=%p now=%ld Timeout=%ld"        , Meta, now, Meta->Timeout );
	D9 ( "Meta=%p fdServer->rx->len=%ld %p %p", Meta, fdServer->rx->len, &fdServer->rx, fdServer );
	D9 ( "Meta=%p Open=%ld Close=%ld"         , Meta, fdServer->fdOpen, fdServer->fdClose );
	D9 ( "Meta=%p Open=%ld Close=%ld"         , Meta, fdClient->fdOpen, fdClient->fdClose );

	if ( ! ( Meta->Status & StatusServerOpen ) ) continue;
	if ( ! ( Meta->Status & StatusClientOpen ) ) continue;

	if ( ! ( Meta->Status & StatusConfigRequest ) ) {
	    Meta->Status     |= StatusConfigRequest;
	    D7 ( "Meta=%p StatusConfigRequest", Meta );
	char                  txCommand_data[1024] = "";
	    sprintf ( (char *)txCommand_data, "%sGET / HTTP/1.0\r\n"        , (char *)txCommand_data             );
	    sprintf ( (char *)txCommand_data, "%sUser-Agent: %s\r\n"        , (char *)txCommand_data, UserAgent  );
	    sprintf ( (char *)txCommand_data, "%sHost: %s\r\n"              , (char *)txCommand_data, RemoteHost );
	    sprintf ( (char *)txCommand_data, "%sConnection: Keep-Alive\r\n", (char *)txCommand_data             );
	    sprintf ( (char *)txCommand_data, "%s\r\n"                      , (char *)txCommand_data             );
	    fdClient->tx->Clean  ();
	    fdClient->tx->Append ( txCommand_data );
	}//if ( ! ( Meta->Status & StatusConfigRequest ) )

	if ( ! ( Meta->Status & StatusConfig ) && ( Meta->Status & StatusConfigRequest ) && ( fdClient->rx->len ) ) {
	    Meta->Status     |= StatusConfig;
	    D7 ( "Meta=%p StatusConfig", Meta );
	//2do host - database - user - match from response
	    //fdClient->rx->ShowString ("fdClient->rx");//bug4debug
	std::vector<std::string> vsConfigRequest = fwSplit( (char *)fdClient->rx->data, '\n' );
	    for ( unsigned long vsConfigRequest_Pos = 0; vsConfigRequest_Pos < vsConfigRequest.size(); vsConfigRequest_Pos++ ) {
	    std::string KeyVal = vsConfigRequest[vsConfigRequest_Pos];
		D9 ( "Meta=%p KeyVal=%s", Meta, KeyVal.c_str() );
		if ( strncmp ( KeyVal.c_str(), ConnectRequest, strlen ( ConnectRequest ) ) == 0 ) {
		    fwReplace ( KeyVal, "\r", "" );
		    fwReplace ( KeyVal, "\n", "" );
		    D9 ( "Meta=%p KeyVal=~%s~", Meta, KeyVal.c_str() );
		    if ( Meta->CONNECT ) free ( Meta->CONNECT );
		    Meta->CONNECT = strdup ( KeyVal.c_str() + strlen ( ConnectRequest ) + 2 );
		    break;
		}
	    }//for ( unsigned long vsConfigRequest_Pos = 0; vsConfigRequest_Pos < vsConfigRequest.size(); vsConfigRequest_Pos++ )
	    fdClient->rx->Free ();
	    if ( ! Meta->CONNECT ) {
		Meta->soLocal->fdClose = 1;
		PEC ( "Meta=%p ! CONNECT", Meta );
	    }
	    D2 ( "Meta=%p CONNECT=~%s~", Meta, Meta->CONNECT );
	}//if (   ( Meta->Status & StatusConfigRequest ) && ( fdClient->rx.len ) )

    //CONNECT HTTP Request
	if ( ! ( Meta->Status & StatusConnectRequest ) && ( Meta->CONNECT ) ) {
	    Meta->Status     |= StatusConnectRequest;
	    D7 ( "Meta=%p StatusConnectRequest", Meta );
	char                  txCommand_data[1024] = "";
	    sprintf ( (char *)txCommand_data, "%s%s\r\n"                          , (char *)txCommand_data, Meta->CONNECT  );
	    sprintf ( (char *)txCommand_data, "%sUser-Agent: %s\r\n"              , (char *)txCommand_data, UserAgent      );
	    sprintf ( (char *)txCommand_data, "%sHost: %s\r\n"                    , (char *)txCommand_data, RemoteHost     );
	    //rintf ( (char *)txCommand_data, "%sHost: http://%s:80\r\n"          , (char *)txCommand_data, RemoteHost     );//2do check 4 virtual-host
	    sprintf ( (char *)txCommand_data, "%sProxy-Connection: Keep-Alive\r\n", (char *)txCommand_data                 );
	    sprintf ( (char *)txCommand_data, "%s\r\n"                            , (char *)txCommand_data                 );
	    fdClient->tx->Clean  ();
	    fdClient->tx->Append ( txCommand_data );
	}//if ( ! ( Meta->Status & StatusConnectRequest ) && ( Meta->CONNECT ) )
    //CONNECT HTTP Request done

    //CONNECT HTTP Response
	if ( ! ( Meta->Status & StatusConnect ) && ( Meta->Status & StatusConnectRequest ) && ( fdClient->rx->len ) ) {
	    Meta->Status     |= StatusConnect;
	    D7 ( "Meta=%p StatusConnect", Meta );

	_bnBlob blRX ( fdClient->rx );
	    //blRX.Save ( "/tmp/blRX.Save.bin" );
	    if (
		   ( strncmp ( (char *)blRX.data, "HTTP/1.0 200", strlen ( "HTTP/1.0 200" ) ) == 0 )
		|| ( strncmp ( (char *)blRX.data, "HTTP/1.1 101", strlen ( "HTTP/1.1 101" ) ) == 0 )
	    ) {
	    unsigned long blRX_Offset = 0;
		for ( blRX_Offset = 0; blRX_Offset < blRX.len; blRX_Offset++ ) {
		    if ( strncmp ( (char *)(blRX.data + blRX_Offset), "\r\n\r\n", strlen ( "\r\n\r\n" ) ) == 0 ) {
			blRX_Offset += 4;
			break;
		    }
		}//for ( blRX_Offset = 0; blRX_Offset < blRX.len; blRX_Offset++ )
		D9 ( "Meta=%p blRX_Offset=%ld", Meta, blRX_Offset );
		fdClient->rx->Clean ();
		fdClient->rx->len = blRX.len - blRX_Offset;
		memcpy ( fdClient->rx->data, blRX.data + blRX_Offset, fdClient->rx->len );
		//fdClient->rx.Save ( "/tmp/fdClientrx.Save.bin" );
		Meta->Status |= StatusConnectedRequest;
	    }//if ( strncmp ( (char *)blRX.data, "HTTP/1.0 200", strlen ( "HTTP/1.0 200" ) ) == 0 )
	    else {
		Meta->soLocal->fdClose = 1;
		PEC ( "Meta=%p ! HTTP/1.0 200", Meta );
	    }
	}//if ( ! ( Meta->Status & StatusConnect ) && ( Meta->Status & StatusConnectRequest ) && ( fdClient->rx.len ) )
    //CONNECT HTTP Response done

	if ( ! ( Meta->Status & StatusConnected ) && ( Meta->Status & StatusConnectedRequest ) ) {
	    Meta->Status     |= StatusConnected;
	    D2 ( "Meta=%p StatusConnected", Meta );//D2
	    bn_delete ( Meta->RemoteTX );
	    fdClient->tx = fdServer->rx;
	    bn_delete ( Meta->LocalTX  );
	    fdServer->tx = fdClient->rx;
	    Meta->Timeout = 0;
	}//if ( ! ( Meta->Status & StatusConnected ) && ( Meta->Status & StatusConnectedRequest ) )

	     fdClient->Error[0] = 0;
	     fdClient->Event    = 0;
	     fdClient->Loop ();
	if ( fdClient->Event ) uSleep = uSleepMin;

	D9 ( "Meta=%p Meta->Status=0x%08X", Meta, Meta->Status );

    }//for ( int vs_stack_pos = 0; vs_stack_pos < (*(*soServer).vs_stack).size(); vs_stack_pos++ )

    D9 ( "(*soServer)=%p", &(*soServer) );
         (*soServer).Error[0] = 0;
         (*soServer).Event    = 0;
         (*soServer).Loop ();
    if ( (*soServer).Event ) uSleep = uSleepMin;

    //uSleepMax = 999999;	//bug4debug
    //uSleepMin = uSleepMax -10;//bug4debug
//uSleep
    if ( !  uSleepMin          ) return NULL;
    if ( !  uSleep             ) uSleep = uSleepMin;
    if (    uSleep > uSleepMax ) uSleep = uSleepMax;
    //D9 ( "uSleep=%ld", uSleep );
    usleep ( uSleep );
             uSleep += uSleepMin;

    return NULL;

}//char *_bnForward::Loop ( void )


char *_bnForward::Close ( void ) {
    D9 ( "this=%p", this );

    (*soServer).Close ();
    InitDone = 0;

    return NULL;

}//char *_bnForward::Close ( void )


//#define bnForward_MAIN
#ifdef bnForward_MAIN
int main ( int argc, char *argv[] ) {
char  Error[1024];
    //albus_log_level = 9;//bug4debug
    //albus_log_file  = strdup ("/tmp/bnForwardd2.log");//bug4debug

string sConfig;
    if ( 2 == argc ) {
    string xConfig = (string)argv[1];
	sConfig = File2String ( xConfig );
    }
    else {
    _bnBlob blSRC;
	blSRC.FileRead ( "stdin" );
	sConfig = (char *)blSRC.data;
    }
    D9 ( "sConfig=~%s~", sConfig.c_str() );

_bnForward bnForward;
vector<_KeyIdxVal> KeyIdxVal = String2KeyIdxVal ( sConfig );
    D9 ( "KeyIdxVal.size()=%d", KeyIdxVal.size() );
int KeyIdxValPos = 0;
    while ( KeyIdxValPos < KeyIdxVal.size() ) {
	D9 ( "KeyIdxVal[%d].Key=~%s~ Val=~%s~", KeyIdxValPos, KeyIdxVal[KeyIdxValPos].Key.c_str(), KeyIdxVal[KeyIdxValPos].Val.c_str() );
    _KeyIdxVal SubKIV = _KeyIdxVal ( KeyIdxVal[KeyIdxValPos].Val );
	if ( 1 == 0 ) {}
	else if ( "log"       == KeyIdxVal[KeyIdxValPos].Key ) albus_log_set ( SubKIV.Key.c_str(), SubKIV.Val.c_str() );
	else if ( "bnForward" == KeyIdxVal[KeyIdxValPos].Key ) bnForward.Set ( KeyIdxVal[KeyIdxValPos].Idx,                  SubKIV.Key, SubKIV.Idx,                  SubKIV.Val );
	else                                                   bnForward.Set ( KeyIdxVal[KeyIdxValPos].Idx, KeyIdxVal[KeyIdxValPos].Key, SubKIV.Idx, KeyIdxVal[KeyIdxValPos].Val );
	KeyIdxValPos++;
    }//while ( KeyIdxValPos < KeyIdxVal.size() )

    //2do pid_t nPid = fork();
    //2do     D8 ( "nPid=%ld getpid=%ld", nPid, getpid () );
    //2do     if ( nPid ) return 0;

long PPID = 0;
#ifndef WIN32
//bug @ WIN32
    PPID = getppid ();
#endif
long loop = 0;
    while ( 1 ) {
	D9 ( "while loop=%ld", loop++ );

	bnForward.Loop ();
	if ( bnForward.Error[0] ) {
	    ER ( "%s", bnForward.Error );
	    break;
	}

    //while ppid is aliave
	//2do #ifndef WIN32
	//2do 	if ( PPID != getppid () ) {
	//2do 	    D2 ( "getpid=%ld PPID=%ld break", getpid (), PPID );
	//2do 	    break;
	//2do 	}
	//2do #endif
	//sleep ( 1 );//bug4debug
	//break;      //bug4debug
    }//while ( 1 )

    D2 ( "exit" );

    return 0;

}//int main ( int argc, char *argv[] )
#endif
