//Copyright 2017 blacknoise.at
//
//   Licensed under the Apache License, Version 2.0 (the "License");
//   you may not use this file except in compliance with the License.
//   You may obtain a copy of the License at
//
//       http://www.apache.org/licenses/LICENSE-2.0
//
//   Unless required by applicable law or agreed to in writing, software
//   distributed under the License is distributed on an "AS IS" BASIS,
//   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//   See the License for the specific language governing permissions and
//   limitations under the License.
//
#ifndef bncSQLDataBase_H
#define bncSQLDataBase_H
#include <string.h>

#include <bncsql.h>
#ifdef bncSQL_DEBUG
# define bncSQLDataBase_DEBUG
#endif
#ifdef bncSQLDataBase_DEBUG
# include <albus_log.h>
#endif

#ifdef WIN32
# include <direct.h>
# include <windows.h>
#else
# include <dlfcn.h>
#endif


extern char bncSQLError[];


char *DriverNotLoad ( void * );


class _bncDataBase {
private:
    void *DriverOpen  ( void );
    void *DriverClose ( void );
    char  DriverName[255];
#ifdef WIN32
HINSTANCE DriverHandle;
#else
    void *DriverHandle;
#endif
    long *Driver_API_Version;
    char *Driver_Version;

public:
//functions in driver
    char *(*dbOpen)         ( void * );
    char *(*dbClose)        ( void * );
    char *(*tbOpen)         ( void * );
    char *(*tbColumnInsert) ( void *, _bncColumn *co, unsigned long );
    char *(*tbClose)        ( void * );
    char *(*tbRecord)       ( void * );
    char *(*tbColumn)       ( void *, const char *, unsigned char **, unsigned long * );
    char *(*tbSQLBuild)     ( void * );
    char *(*coDEF)          ( void * );

    void *Config     ( int argc, char **argv );
    void *Show       ( void );
    void *Open       ( void );
    void *Destructor ( void );

    char  Error[255];
    char *Host;
    char *Base;
    char *User;
    char *Pass;
    char *Driver;
    long  Port;
    char *Options;

    char *TLSCertificate;
    char *TLSPrivateKey;
    char *TLSPrivateKeyPass;
    char *TLSFingerprint00;
    char *TLSFingerprint01;
    char *TLSFingerprint02;
    char *TLSFingerprint03;
    char *TLSFingerprint04;

    char *LibPath;
    char *Home;
    char  Quote;
    char *lc_all;
    void *Conn; //Connection 2 db
    void *bncSQLProxyVP;
    char *Encoding; //Encoding = strdup ( "UTF8" ); 2do
    char *Builtin ( void );

    _bncDataBase() {
#ifdef  bncSQLDataBase_DEBUG
	D7 ( "this=%p", this );//best D7
#endif
	memset ( this, 0x00, sizeof ( _bncDataBase ) );
	Quote  = ' '          ;
	lc_all = strdup ( "" );
	*(void **)(&dbOpen)         = (void *)&DriverNotLoad;
	*(void **)(&dbClose)        = (void *)&DriverNotLoad;
	*(void **)(&tbOpen)         = (void *)&DriverNotLoad;
	*(void **)(&tbColumnInsert) = (void *)&DriverNotLoad;
	*(void **)(&tbClose)        = (void *)&DriverNotLoad;
	*(void **)(&tbRecord)       = (void *)&DriverNotLoad;
	*(void **)(&tbColumn)       = (void *)&DriverNotLoad;
	*(void **)(&tbSQLBuild)     = (void *)&DriverNotLoad;
	*(void **)(&coDEF)          = (void *)&DriverNotLoad;
    }

    ~_bncDataBase() {
#ifdef  bncSQLDataBase_DEBUG
	D7 ( "this=%p", this );//best D7
#endif
	Destructor ();
	bn_free ( Host              );
	bn_free ( Base              );
	bn_free ( User              );
	bn_free ( Pass              );
	bn_free ( Driver            );
	bn_free ( Options           );
	bn_free ( TLSCertificate    );
	bn_free ( TLSPrivateKey     );
	bn_free ( TLSPrivateKeyPass );
	bn_free ( TLSFingerprint00  );
	bn_free ( TLSFingerprint01  );
	bn_free ( TLSFingerprint02  );
	bn_free ( TLSFingerprint03  );
	bn_free ( TLSFingerprint04  );
	bn_free ( LibPath           );
	bn_free ( Home              );
	bn_free ( lc_all            );
	//_free ( Conn    ); done via driver
    }

    void *Extern;

};

#endif
