//Copyright 2017 blacknoise.at
//
//   Licensed under the Apache License, Version 2.0 (the "License");
//   you may not use this file except in compliance with the License.
//   You may obtain a copy of the License at
//
//       http://www.apache.org/licenses/LICENSE-2.0
//
//   Unless required by applicable law or agreed to in writing, software
//   distributed under the License is distributed on an "AS IS" BASIS,
//   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//   See the License for the specific language governing permissions and
//   limitations under the License.
//
#ifndef bncSQLProxy_H
#define bncSQLProxy_H
#include <string.h>
#ifdef __APPLE__
# include <malloc/malloc.h>
#else
# include <malloc.h>
#endif
#include <signal.h>
#include <time.h>
#ifdef WIN32
//# include <direct.h>
//# include <windows.h>
#else
//# include <dlfcn.h>
#endif

#ifdef bncSQL_DEBUG
# define bncSQLProxy_DEBUG
#endif
#ifdef bncSQLProxy_DEBUG
# include <albus_log.h>
#endif

#include <bnForward.h>

#ifndef bn_free
#define bn_free(a) {\
    if (a) { free(a); }\
        a = NULL;\
}
#endif


class _bncSQLProxy {
private:
    time_t          Timeout;
    unsigned long   uSleep;
    unsigned long   uSleepMin;
    unsigned long   uSleepMax;

public:
    void *Open       ( void );
    void *Show       ( void );
    void *Status     ( void );
    void *Close      ( void );
    void *Destructor ( void );

    unsigned long   pThreadIndex;
    int             pThreadExit;

    char            Error[1024];
    char           *RemoteHost;
    unsigned short  RemoteTLS;
    char           *RemoteTLSCertificate;
    char           *RemoteTLSPrivateKey;
    char           *RemoteTLSPrivateKeyPass;
    char           *RemoteTLSFingerprint00;
    char           *RemoteTLSFingerprint01;
    char           *RemoteTLSFingerprint02;
    char           *RemoteTLSFingerprint03;
    char           *RemoteTLSFingerprint04;

    char           *LocalHost;
    unsigned short  LocalPort;
    char           *Agent;
    char           *AgentVersion;
    char           *Driver;
    _bnForward     *bnForward;

    _bncSQLProxy() {
#ifdef  bncSQLProxy_DEBUG
	D9 ( "this=%p", this );
#endif
	memset ( this, 0x00, sizeof ( _bncSQLProxy ) );
	uSleepMin         =   1000;
	uSleepMax         = 200000;
    }

    ~_bncSQLProxy() {
#ifdef  bncSQLProxy_DEBUG
	D9 ( "this=%p", this );
#endif
	Close ();
	bn_free ( RemoteHost   );
	bn_free ( LocalHost    );
	bn_free ( Agent        );
	bn_free ( AgentVersion );
	bn_free ( Driver       );
    }

};

#endif
