//Copyright 2017-2018 blacknoise.at
//
//   Licensed under the Apache License, Version 2.0 (the "License");
//   you may not use this file except in compliance with the License.
//   You may obtain a copy of the License at
//
//       http://www.apache.org/licenses/LICENSE-2.0
//
//   Unless required by applicable law or agreed to in writing, software
//   distributed under the License is distributed on an "AS IS" BASIS,
//   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//   See the License for the specific language governing permissions and
//   limitations under the License.
//
#ifdef HAVE_CONFIG
# include <config.h>
#endif
#ifdef bnSSL_ENABLED
#include "bnSSL.h"
#include <stdio.h>
#include <string.h>

//#define HAVE_SSL
#ifdef  HAVE_SSL
# include <openssl/ssl.h>
# include <openssl/err.h>

# include <openssl/pem.h>
# include <openssl/rand.h>
# include <openssl/rsa.h>
# include <openssl/crypto.h>
# include <openssl/bn.h>
# include <openssl/x509v3.h>
#endif

#include <base64.h>

#include <bnRC4.h>
#include <bnSplitPos.h>
#include <albus_log.h>
#ifndef bnSSL_DEBUG
# undef  MI
# define MI
# undef  D8
# define D8
# undef  D9
# define D9
#endif


#define Error MainError
extern char Error[];


#ifdef  HAVE_SSL
typedef struct pw_cb_data {
    const void *password;
    const char *prompt_info;
} PW_CB_DATA;


static int password_callback2 ( char *buf, int bufsiz, int verify, PW_CB_DATA *cb_tmp ) {
int res = 0;
const char *password = NULL;
PW_CB_DATA *cb_data = (PW_CB_DATA *)cb_tmp;

    if (cb_data) {
	if (cb_data->password) password = (const char *)cb_data->password;
    }

    if (password) {
	res = strlen(password);
	if (res > bufsiz) res = bufsiz;
	memcpy(buf, password, res);
	return res;
    }

    return res;

}//static int password_callback2 ( char *buf, int bufsiz, int verify, PW_CB_DATA *cb_tmp )
#endif


char *bnSSLGenRSAKey ( char *rsafile, char *rsapasswd, int rsabits ) {
    D9 ( "call" );
#ifdef  HAVE_SSL
char   *rsafileTX = NULL;
FILE   *rsaFD     = NULL;
BIGNUM *e         = NULL;
RSA    *rsa       = NULL;
const EVP_CIPHER *enc = NULL;
long  rsapasswdLen = 0;

    OpenSSL_add_all_algorithms ();
    OpenSSL_add_all_ciphers    ();
    ERR_load_crypto_strings    ();

    if ( ! rsafile ) SEGT ( "! rsafile" );
    if ( ! rsabits ) rsabits = 2048;

    D2 ( "rsafile=~%s~ rsabits=%d", rsafile, rsabits );
    D9 ( "rsapasswd=~%s~", rsapasswd );

    rsaFD = fopen ( rsafile, "rb" );//dont overwrite
    if ( rsaFD ) SEGT ( "file %s already exsist", rsafile );

    e = BN_new (); if ( ! e ) SEGT ( "! e" );
    rsa = RSA_new_method ( NULL ); if ( ! rsa ) SEGT ( "! rsa" );
    if ( ! BN_set_word ( e, RSA_F4 ) ) SEGT ( "! BN_set_word ( e, RSA_F4 )" );
    if ( ! RSA_generate_key_ex ( rsa, rsabits, e, NULL ) ) SEGT ( "! RSA_generate_key_ex ( rsa, bits, e, NULL )" );

    //passwd = strdup ("dummy");//bug4debug
    if ( rsapasswd ) {
	enc          = EVP_aes_256_cbc ();
	rsapasswdLen = strlen ( rsapasswd );
    }

    rsaFD = fopen ( rsafile, "wb" ); if ( ! rsaFD ) SEGT ( "! fopen %s", rsafile );
    PEM_write_RSAPrivateKey ( rsaFD, rsa, enc, (unsigned char *)rsapasswd, rsapasswdLen, NULL, NULL );
    D2 ( "rsafile=~%s~", rsafile );

    rsafileTX = strdup ( rsafile );

my_return:
    D9 ( "my_return" );
    if ( rsaFD ) fclose   ( rsaFD );
    if ( e     ) BN_free  ( e     );
    if ( rsa   ) RSA_free ( rsa   );
    return rsafileTX;
#else
    SER ( "no ssl build" );
#endif
}//char *bnSSLGenRSAKey ( char *rsafile, char *rsapasswd, int rsabits )


char *bnSSLGenCRT ( char *crtfile, char *crtconfig, char *rsafile, char *rsapasswd, char *cacrtfile, char *carsafile, char *carsapasswd ) {
    D9 ( "call" );
#ifdef  HAVE_SSL
char       *crtfileTX = NULL;
FILE       *rsaFD     = NULL;
RSA        *rsa       = NULL;
RSA        *rsaRX     = NULL;
EVP_PKEY   *pkey      = NULL;
X509       *subject   = NULL;
X509       *issuer    = NULL; // Root CA | = subject if self signed
X509       *issuerRX  = NULL;
X509_NAME  *name      = NULL;
X509_NAME  *caname    = NULL;
X509V3_CTX *ctx       = NULL;
X509_REQ   *req       = NULL;
X509_CRL   *crl       = NULL;
int         flags     =    0;
X509_EXTENSION *ex    = NULL;
FILE       *crtFD     = NULL;
FILE       *cacrtFD   = NULL;
FILE       *carsaFD   = NULL;
RSA        *carsa     = NULL;
RSA        *carsaRX   = NULL;
EVP_PKEY   *capkey    = NULL;
EVP_CIPHER *enc       = NULL;

_bnSplitPos *spCfgLine = NULL;
long  default_days           = 9000;//  365;//default
char *commonName             = NULL;
char *organizationalUnitName = NULL;
char *organizationName       = NULL;
char *localityName           = NULL;
char *stateOrProvinceName    = NULL;
char *countryName            = NULL;
char *emailAddress           = NULL;
char *subjectAltName         = NULL;
char *subjectKeyIdentifier   = NULL;
char *authorityKeyIdentifier = NULL;
char *basicConstraints       = NULL;
char *keyUsage               = NULL;
char *extendedKeyUsage       = NULL;

    OpenSSL_add_all_algorithms ();
    OpenSSL_add_all_ciphers    ();
    ERR_load_crypto_strings    ();

    if ( ! rsafile ) SEGT ( "! rsafile" );
    if ( ! crtfile ) SEGT ( "! crtfile" );

    D9 ( "crtconfig=~%s~", crtconfig );

    spCfgLine = bnSplitPosNew ( crtconfig, strlen(crtconfig) );
    spCfgLine->Delimiter = '\n';
    spCfgLine->Quote     = NULL; //unset default
    spCfgLine->Ignore    = NULL; //unset default
    while ( bnSplitPosNext ( spCfgLine ) ) {
    _bnSplitPos *spKeyVal = bnSplitPosNew ( spCfgLine->Input + spCfgLine->Beg, spCfgLine->Len );
	spKeyVal->Delimiter = '=';
	bnSplitPosIndex ( spKeyVal );
	D9 ( "spKeyVal->Records=%ld", spKeyVal->Records );
	if ( 2 == spKeyVal->Records ) {
	char *Key = bnSplitPosGet ( spKeyVal, 0 );
	char *Val = bnSplitPosGet ( spKeyVal, 1 );
	    D9 ( "Key=~%s~ Val=~%s~", Key, Val );
	    if ( strcmp ( "default_days"           , Key ) == 0 ) default_days           = atol   ( Val );
	    if ( strcmp ( "commonName"             , Key ) == 0 ) commonName             = strdup ( Val );
	    if ( strcmp ( "organizationalUnitName" , Key ) == 0 ) organizationalUnitName = strdup ( Val );
	    if ( strcmp ( "organizationName"       , Key ) == 0 ) organizationName       = strdup ( Val );
	    if ( strcmp ( "localityName"           , Key ) == 0 ) localityName           = strdup ( Val );
	    if ( strcmp ( "stateOrProvinceName"    , Key ) == 0 ) stateOrProvinceName    = strdup ( Val );
	    if ( strcmp ( "countryName"            , Key ) == 0 ) countryName            = strdup ( Val );
	    if ( strcmp ( "emailAddress"           , Key ) == 0 ) emailAddress           = strdup ( Val );
	    if ( strcmp ( "subjectAltName"         , Key ) == 0 ) subjectAltName         = strdup ( Val );
	    if ( strcmp ( "subjectKeyIdentifier"   , Key ) == 0 ) subjectKeyIdentifier   = strdup ( Val );
	    if ( strcmp ( "authorityKeyIdentifier" , Key ) == 0 ) authorityKeyIdentifier = strdup ( Val );
	    if ( strcmp ( "basicConstraints"       , Key ) == 0 ) basicConstraints       = strdup ( Val );
	    if ( strcmp ( "keyUsage"               , Key ) == 0 ) keyUsage               = strdup ( Val );
	    if ( strcmp ( "extendedKeyUsage"       , Key ) == 0 ) extendedKeyUsage       = strdup ( Val );
	    if ( Key ) free ( Key );
	    if ( Val ) free ( Val );
	}//if ( 2 == spKeyVal->Records )
	bnSplitPosFree ( spKeyVal );
    }//while ( bnSplitPosNext ( bnSplitPos ) )

    if ( ! organizationName       ) organizationName       = strdup ( "Internet Widgits Pty Ltd" );//default
    if ( ! stateOrProvinceName    ) stateOrProvinceName    = strdup ( "Some-State"               );//default
    if ( ! countryName            ) countryName            = strdup ( "AU"                       );//default
    if ( ! subjectKeyIdentifier   ) subjectKeyIdentifier   = strdup ( "hash"                     );//default
    if ( ! authorityKeyIdentifier ) authorityKeyIdentifier = strdup ( "keyid:always"             );//default
    if ( ! basicConstraints       ) basicConstraints       = strdup ( "CA:TRUE"                  );//default

    if ( emailAddress ) {
    char myBuffer[1024];
	if ( commonName ) {
	    sprintf ( myBuffer, "%s/%s", commonName, emailAddress );
	    free ( commonName );
	    commonName = strdup ( myBuffer );
	}
	else {
	    sprintf ( myBuffer, "%s/%s", organizationName, emailAddress );
	    free ( organizationName );
	    organizationName = strdup ( myBuffer );
	}
    }//if ( emailAddress )

    D2 ( "default_days          = %ld" , default_days           );
    D2 ( "commonName            = ~%s~", commonName             );
    D2 ( "organizationalUnitName= ~%s~", organizationalUnitName );
    D2 ( "organizationName      = ~%s~", organizationName       );
    D2 ( "localityName          = ~%s~", localityName           );
    D2 ( "stateOrProvinceName   = ~%s~", stateOrProvinceName    );
    D2 ( "countryName           = ~%s~", countryName            );
    D2 ( "emailAddress          = ~%s~", emailAddress           );
    D2 ( "subjectAltName        = ~%s~", subjectAltName         );
    D2 ( "subjectKeyIdentifier  = ~%s~", subjectKeyIdentifier   );
    D2 ( "authorityKeyIdentifier= ~%s~", authorityKeyIdentifier );
    D2 ( "basicConstraints      = ~%s~", basicConstraints       );
    D2 ( "keyUsage              = ~%s~", keyUsage               );
    D2 ( "extendedKeyUsage      = ~%s~", extendedKeyUsage       );

PW_CB_DATA cb_data;
    cb_data.password = rsapasswd;

    rsaFD = fopen ( rsafile, "rb" ); if ( ! rsaFD ) SEGT ( "! fopen %s", rsafile );
    pkey = PEM_read_PrivateKey ( rsaFD, (EVP_PKEY **)&enc, (pem_password_cb *)password_callback2, &cb_data );
    if ( ! pkey ) SEGT ( "! pkey" );
    D9 ( "pkey=%p", pkey );


//Certificate
    subject = X509_new (); if ( ! subject ) SEGT ( "! subject" );

    if ( cacrtfile ) {
	cacrtFD = fopen ( cacrtfile, "rb" ); if ( ! cacrtFD ) SEGT ( "! fopen %s", rsafile );
	issuerRX = PEM_read_X509 ( cacrtFD, &issuer, NULL, NULL ); if ( ! issuerRX ) SEGT ( "! issuerRX" );
	carsaFD = fopen ( carsafile, "rb" ); if ( ! carsaFD ) SEGT ( "! fopen %s", carsafile );
    EVP_CIPHER *enc2       = NULL;
    PW_CB_DATA cb_data2;
	cb_data2.password = carsapasswd;
	capkey = PEM_read_PrivateKey ( carsaFD, (EVP_PKEY **)&enc2, (pem_password_cb *)password_callback2, &cb_data2 );
	if ( ! capkey ) SEGT ( "! capkey" );
	D9 ( "capkey=%p", capkey );
    }//if ( cacrtfile )
    else {
	issuer  = subject; //self signed
	capkey  = pkey;    //self signed
    }
    if ( ! issuer ) SEGT ( "! issuer" );
    if ( ! capkey ) SEGT ( "! capkey" );

    X509_set_version ( subject, 2 );
    ASN1_INTEGER_set ( X509_get_serialNumber ( subject ),    0x02 ); //bug2
    X509_gmtime_adj  ( X509_get_notBefore    ( subject ),       0 );
    X509_gmtime_adj  ( X509_get_notAfter     ( subject ), default_days * 24 * 3600 );
    X509_set_pubkey  ( subject, pkey );

    name = X509_get_subject_name ( subject ); if ( ! name ) SEGT ( "! name" );

    if ( countryName            ) X509_NAME_add_entry_by_txt ( name, "C" , MBSTRING_ASC, (unsigned char *)countryName           , -1, -1, 0 );
    if ( stateOrProvinceName    ) X509_NAME_add_entry_by_txt ( name, "ST", MBSTRING_ASC, (unsigned char *)stateOrProvinceName   , -1, -1, 0 );
    if ( organizationName       ) X509_NAME_add_entry_by_txt ( name, "O" , MBSTRING_ASC, (unsigned char *)organizationName      , -1, -1, 0 );
    if ( commonName             ) X509_NAME_add_entry_by_txt ( name, "CN", MBSTRING_ASC, (unsigned char *)commonName            , -1, -1, 0 );
    if ( organizationalUnitName ) X509_NAME_add_entry_by_txt ( name, "OU", MBSTRING_ASC, (unsigned char *)organizationalUnitName, -1, -1, 0 );
    if ( localityName           ) X509_NAME_add_entry_by_txt ( name, "L" , MBSTRING_ASC, (unsigned char *)localityName          , -1, -1, 0 );

    caname = X509_get_subject_name ( issuer  ); if ( ! name ) SEGT ( "! name" );
    X509_set_issuer_name ( subject, caname );
    X509_sign ( subject, capkey, EVP_sha256() );

//X509_EXTENSION
    ctx = (X509V3_CTX *)malloc ( sizeof ( X509V3_CTX ) ); if ( ! ctx ) SEGT ( "! ctx" );
    X509V3_set_ctx_nodb ( ctx );
    X509V3_set_ctx ( ctx, issuer, subject, req, crl, flags );

    if ( subjectKeyIdentifier ) {
	ex = X509V3_EXT_conf_nid ( NULL, ctx, NID_subject_key_identifier  , subjectKeyIdentifier   ); if ( ! ex ) SEGT ( "! ex" );
	X509_add_ext ( subject, ex, -1 );
	X509_EXTENSION_free ( ex ); ex = NULL;
    }

    if ( authorityKeyIdentifier ) {
	ex = X509V3_EXT_conf_nid ( NULL, ctx, NID_authority_key_identifier, authorityKeyIdentifier ); if ( ! ex ) SEGT ( "! ex" );
	X509_add_ext ( subject, ex, -1 );
	X509_EXTENSION_free ( ex ); ex = NULL;
    }
    if ( basicConstraints       ) {
	ex = X509V3_EXT_conf_nid ( NULL, ctx, NID_basic_constraints       , basicConstraints       ); if ( ! ex ) SEGT ( "! ex" );
	X509_add_ext ( subject, ex, -1 );
	X509_EXTENSION_free ( ex ); ex = NULL;
    }

    if ( keyUsage ) {
	ex = X509V3_EXT_conf_nid ( NULL, ctx, NID_key_usage               , keyUsage               ); if ( ! ex ) SEGT ( "! ex" );
	X509_add_ext ( subject, ex, -1 );
	X509_EXTENSION_free ( ex ); ex = NULL;
    }

    if ( extendedKeyUsage ) {
	ex = X509V3_EXT_conf_nid ( NULL, ctx, NID_ext_key_usage           , extendedKeyUsage       ); if ( ! ex ) SEGT ( "! ex" );
	X509_add_ext ( subject, ex, -1 );
	X509_EXTENSION_free ( ex ); ex = NULL;
    }

    if ( subjectAltName ) {
	ex = X509V3_EXT_conf_nid ( NULL, ctx, NID_subject_alt_name        , subjectAltName         ); if ( ! ex ) SEGT ( "! ex" );
	X509_add_ext ( subject, ex, -1 );
	X509_EXTENSION_free ( ex ); ex = NULL;
    }
//X509_EXTENSION done

    crtFD = fopen ( crtfile, "wb" ); if ( ! crtFD ) SEGT ( "! crtFD=%s", crtfile );
    PEM_write_X509 ( crtFD, subject );

    crtfileTX = strdup ( crtfile );

my_return:
    D9 ( "my_return" );
    if ( issuer == subject ) issuer = NULL;
    if ( capkey == pkey    ) capkey = NULL;
    if ( req     ) X509_REQ_free       ( req     );
    if ( crl     ) X509_CRL_free       ( crl     );
    if ( ex      ) X509_EXTENSION_free ( ex      );
    //if ( name    ) X509_NAME_free      ( name    );//2do double free
    //if ( caname  ) X509_NAME_free      ( caname  );//2do double free
    if ( issuer  ) X509_free           ( issuer  );
    if ( rsaFD   ) fclose              ( rsaFD   );
    if ( rsa     ) RSA_free            ( rsa     );
    if ( ctx     )                free ( ctx     );// X509V3_CTX_free     ( ctx     ); 2do
    if ( subject ) X509_free           ( subject );
    if ( crtFD   ) fclose              ( crtFD   );
    if ( cacrtFD ) fclose              ( cacrtFD );
    if ( carsaFD ) fclose              ( carsaFD );
    if ( carsa   ) RSA_free            ( carsa   );
    if ( capkey  ) EVP_PKEY_free       ( capkey  );
    if ( pkey    ) EVP_PKEY_free       ( pkey    );
    // ( enc     ) EVP_CIPHER_free     ( enc     );//2do

    bnSplitPosFree ( spCfgLine );
    if ( commonName             ) free ( commonName             );
    if ( organizationalUnitName ) free ( organizationalUnitName );
    if ( organizationName       ) free ( organizationName       );
    if ( localityName           ) free ( localityName           );
    if ( stateOrProvinceName    ) free ( stateOrProvinceName    );
    if ( countryName            ) free ( countryName            );
    if ( emailAddress           ) free ( emailAddress           );
    if ( subjectAltName         ) free ( subjectAltName         );
    if ( subjectKeyIdentifier   ) free ( subjectKeyIdentifier   );
    if ( authorityKeyIdentifier ) free ( authorityKeyIdentifier );
    if ( basicConstraints       ) free ( basicConstraints       );
    if ( keyUsage               ) free ( keyUsage               );
    if ( extendedKeyUsage       ) free ( extendedKeyUsage       );

    return crtfileTX;
#else
    SER ( "no ssl build" );
#endif
}//char *bnSSLGenCRT ( char *crtfile, char *crtconfig, char *rsafile, char *rsapasswd, char *cacrtfile, char *carsafile, char *carsapasswd )


#ifdef HAVE_SSL
_bnBlob *bnSSLBase2Encrypt ( const char *Base64 ) {
    D9 ( "call" );

std::string b2eData = base64_decode ( Base64 );
    D9 ( "b2eData.size()=%ld", b2eData.size() );

_bnBlob *b2eBlob = new _bnBlob ( b2eData.size() );
//memcpy ( b2eBlob->data____, &b2eData[0], b2eData.size() );//2do check @ all OK
    copy ( b2eData.begin(), b2eData.end(), b2eBlob->data );
    b2eBlob->len = b2eBlob->size;
    return b2eBlob;

}//_bnBlob *bnSSLBase2Encrypt ( const char *Base64 )


_bnBlob *bnSSLEncrypt2Sign ( _bnBlob *Encrypt, _bnBlob *Key ) {
//decrypt with private -> client_rsa.key
    D9 ( "call" );

BIO* bio = BIO_new ( BIO_s_mem () );
    BIO_write ( bio, Key->data, Key->len );
    if ( ! bio  ) SER ( "! bio"  );
EVP_PKEY* pkey = 0;
    PEM_read_bio_PrivateKey ( bio, &pkey, 0, 0 );
    BIO_free ( bio );
    if ( ! pkey ) SER ( "! pkey" );
RSA* rsa = EVP_PKEY_get1_RSA ( pkey );
    EVP_PKEY_free ( pkey );
    if ( ! rsa ) SER ( "! rsa" );
    D9 ( "rsa=%p", rsa );

long keysize = RSA_size ( rsa );
    if ( ! keysize ) SER ( "! keysize" );
    D9 ( "keysize=%ld", keysize );

_bnBlob *Sign = new _bnBlob ( Encrypt->len );
_bnBlob  Block ( keysize );
long LoopOffset = 0;
long BlockSize  = keysize;
    while ( LoopOffset < Encrypt->len ) {
	D9 ( "LoopOffset=%08x Encrypt->len=%08x BlockSize=%ld", LoopOffset, Encrypt->len, BlockSize );
    long RSA_private_decryptRX = RSA_private_decrypt ( keysize, Encrypt->data + LoopOffset, Block.data, rsa, RSA_NO_PADDING );
	if ( RSA_private_decryptRX != keysize ) SER ( "RSA_private_decryptRX != keysize" );
	D9 ( "RSA_private_decryptRX=%ld", RSA_private_decryptRX );
	memcpy ( Sign->data + Sign->len, Block.data + 1, BlockSize - 1 );
	Sign->len  += BlockSize - 1;
	LoopOffset += BlockSize;
    }//while ( LoopOffset < Encrypt->len )

    RSA_free ( rsa );
    return Sign;

}//_bnBlob *bnSSLEncrypt2Sign ( _bnBlob *Encrypt, _bnBlob *Key )


_bnBlob *bnSSLSign2Data ( _bnBlob *Sign, _bnBlob *Cert ) {
//decrypt with public -> bnplugin_rsa.crt
    D9 ( "call" );

BIO* bio = BIO_new ( BIO_s_mem () );
    BIO_write ( bio, Cert->data, Cert->len );
    if ( ! bio  ) SER ( "! bio"  );
X509      *x509 = PEM_read_bio_X509 ( bio, NULL, NULL, NULL );
    BIO_free ( bio );
    if ( ! x509 ) SER ( "! x509" );
EVP_PKEY  *pkey = X509_get_pubkey ( x509 );
    X509_free ( x509 );
    if ( ! pkey ) SER ( "! pkey" );
RSA* rsa = EVP_PKEY_get1_RSA ( pkey );
    EVP_PKEY_free ( pkey );
    if ( ! rsa ) SER ( "! rsa" );
    D9 ( "rsa=%p", rsa );

long keysize = RSA_size ( rsa );
    if ( ! keysize ) SER ( "! keysize" );
    D9 ( "keysize=%ld", keysize );

_bnBlob *Data = new _bnBlob ( Sign->len );
_bnBlob  Block ( keysize );
long LoopOffset = 0;
long BlockSize  = keysize;
    while ( LoopOffset < Sign->len ) {
	D9 ( "LoopOffset=%08x Sign->len=%08x BlockSize=%ld", LoopOffset, Sign->len, BlockSize );
    long RSA_public_decryptRX = RSA_public_decrypt ( keysize, Sign->data + LoopOffset, Block.data, rsa, RSA_NO_PADDING );
	if ( RSA_public_decryptRX != keysize ) SER ( "RSA_public_decryptRX != keysize" );
	D9 ( "RSA_public_decryptRX=%ld", RSA_public_decryptRX );
	memcpy ( Data->data + Data->len, Block.data + 1, BlockSize - 1 );
	Data->len  += BlockSize - 1;
	LoopOffset += BlockSize;
    }//while ( LoopOffset < Sign->len )

    RSA_free ( rsa );
    return Data;

}//_bnBlob *bnSSLSign2Data ( _bnBlob *Sign, _bnBlob *Cert )


char *bnSSLEncrypt2Base ( _bnBlob *Encrypt ) {
    D9 ( "call" );

std::string e2bData = base64_encode ( (unsigned char const* )Encrypt->data, (unsigned int)Encrypt->len );
long  BaseLen = e2bData.size () + 1;
char *Base = new char[ BaseLen ];
    memset ( Base, 0x00, BaseLen );
    strcpy ( Base, e2bData.c_str () );

    return Base;
}//char *bnSSLEncrypt2Base ( _bnBlob *Encrypt )


_bnBlob *bnSSLSign2Encrypt ( _bnBlob *Sign, _bnBlob *Cert ) {
//encrypt with public -> client_rsa.crt
    D9 ( "call" );

BIO* bio = BIO_new ( BIO_s_mem () );
    BIO_write ( bio, Cert->data, Cert->len );
    if ( ! bio  ) SER ( "! bio"  );
X509      *x509 = PEM_read_bio_X509 ( bio, NULL, NULL, NULL );
    BIO_free ( bio );
    if ( ! x509 ) SER ( "! x509" );
EVP_PKEY  *pkey = X509_get_pubkey ( x509 );
    X509_free ( x509 );
    if ( ! pkey ) SER ( "! pkey" );
RSA* rsa = EVP_PKEY_get1_RSA ( pkey );
    EVP_PKEY_free ( pkey );
    if ( ! rsa ) SER ( "! rsa" );
    D9 ( "rsa=%p", rsa );

long keysize = RSA_size ( rsa );
    if ( ! keysize ) SER ( "! keysize" );
    D9 ( "keysize=%ld", keysize );

_bnBlob *Encrypt = new _bnBlob ( Sign->len * 2 );
_bnBlob  Block ( keysize );
long LoopOffset = 0;
long BlockSize  = 0;
    while ( LoopOffset < Sign->len ) {
	BlockSize = Sign->len - LoopOffset;
	if ( BlockSize > keysize - 1 ) BlockSize = keysize - 1;
	if ( BlockSize <= 0          ) break;
	D9 ( "LoopOffset=%08x Sign->len=%08x BlockSize=%ld", LoopOffset, Sign->len, BlockSize );
	         Block.Clean ();
	memcpy ( Block.data + 1, Sign->data + LoopOffset, BlockSize );
    long RSA_public_encryptRX = RSA_public_encrypt ( keysize, Block.data, Encrypt->data + Encrypt->len, rsa, RSA_NO_PADDING );
	if ( RSA_public_encryptRX != keysize ) SER ( "RSA_public_encryptRX != keysize" );
	D9 ( "RSA_public_encryptRX=%ld", RSA_public_encryptRX );
	Encrypt->len += RSA_public_encryptRX;
	LoopOffset += BlockSize;
    }//while ( LoopOffset < Data->len )

    RSA_free ( rsa );
    return Encrypt;

}//_bnBlob *bnSSLSign2Encrypt ( _bnBlob *Sign, _bnBlob *Cert )


_bnBlob *bnSSLData2Sign ( _bnBlob *Data, _bnBlob *Key ) {
//encrypt with private -> bnplugin_rsa.key
    D9 ( "call" );

BIO* bio = BIO_new ( BIO_s_mem () );
    BIO_write ( bio, Key->data, Key->len );
    if ( ! bio  ) SER ( "! bio"  );
EVP_PKEY* pkey = 0;
    PEM_read_bio_PrivateKey ( bio, &pkey, 0, 0 );
    BIO_free ( bio );
    if ( ! pkey ) SER ( "! pkey" );
RSA* rsa = EVP_PKEY_get1_RSA ( pkey );
    EVP_PKEY_free ( pkey );
    if ( ! rsa ) SER ( "! rsa" );
    D9 ( "rsa=%p", rsa );

long keysize = RSA_size ( rsa );
    if ( ! keysize ) SER ( "! keysize" );
    D9 ( "keysize=%ld", keysize );

_bnBlob *Sign = new _bnBlob ( Data->len * 2 );
_bnBlob  Block ( keysize );
long LoopOffset = 0;
long BlockSize  = 0;
    while ( LoopOffset < Data->len ) {
	BlockSize = Data->len - LoopOffset;
	if ( BlockSize > keysize - 1 ) BlockSize = keysize - 1;
	if ( BlockSize <= 0          ) break;
	D9 ( "LoopOffset=%08x Data->len=%08x BlockSize=%ld", LoopOffset, Data->len, BlockSize );
	         Block.Clean ();
	memcpy ( Block.data + 1, Data->data + LoopOffset, BlockSize );
    long RSA_private_encryptRX = RSA_private_encrypt ( keysize, Block.data, Sign->data + Sign->len, rsa, RSA_NO_PADDING );
	if ( RSA_private_encryptRX != keysize ) SER ( "RSA_private_encryptRX != keysize" );
	D9 ( "RSA_private_encryptRX=%ld", RSA_private_encryptRX );
	Sign->len += RSA_private_encryptRX;
	LoopOffset += BlockSize;
    }//while ( LoopOffset < Data->len )

    RSA_free ( rsa );
    return Sign;

}//_bnBlob *bnSSLData2Sign ( _bnBlob *Data, _bnBlob *Key )
#endif


_bnBlob *bnSSLEncrypt ( _bnBlob *Data, _bnBlob *Key, const char *password, int base64 ) {
    D9 ( "call" );
#ifdef  HAVE_SSL
    if ( ! Data      ) SER ( "! Data"      );
    if ( ! Data->len ) SER ( "! Data->len" );
    if ( ! Key       ) SER ( "! Key"       );
    if ( ! Key->len  ) SER ( "! Key->len"  );
    if (   password  ) SER ( "password set, bug2 src it out" );
    //Data->Show ( "Data" );//bug4debug
    //Key->ShowString ( "Key" );//bug4debug

    OpenSSL_add_all_algorithms ();
    OpenSSL_add_all_ciphers    ();
    ERR_load_crypto_strings    ();

_bnBlob *blEncrypt = NULL;
char *TLSPrivateKey = (char *)Key->data;
    if      ( ( strlen ( TLSPrivateKey ) > strlen ( "-----BEGIN RSA PRIVATE KEY-----" ) ) && ( strncmp ( TLSPrivateKey, "-----BEGIN RSA PRIVATE KEY-----", strlen ( "-----BEGIN RSA PRIVATE KEY-----" ) ) == 0 ) ) {
	blEncrypt = bnSSLData2Sign    ( Data, Key ); //encrypt with private -> my_rsa.key
    }
    else if ( ( strlen ( TLSPrivateKey ) > strlen ( "-----BEGIN PRIVATE KEY-----"     ) ) && ( strncmp ( TLSPrivateKey, "-----BEGIN PRIVATE KEY-----"    , strlen ( "-----BEGIN PRIVATE KEY-----"     ) ) == 0 ) ) {
	blEncrypt = bnSSLData2Sign    ( Data, Key ); //encrypt with private -> my_rsa.key
    }
    else if ( ( strlen ( TLSPrivateKey ) > strlen ( "-----BEGIN CERTIFICATE-----"     ) ) && ( strncmp ( TLSPrivateKey, "-----BEGIN CERTIFICATE-----"    , strlen ( "-----BEGIN CERTIFICATE-----"     ) ) == 0 ) ) {
	blEncrypt = bnSSLSign2Encrypt ( Data, Key ); //encrypt with public -> client_rsa.crt
    }
    if ( ! blEncrypt ) SER ( "! blEncrypt" );
    //blEncrypt->Show ( "blEncrypt" );//bug4debug

    if ( base64 ) {
    char *chBase64 = bnSSLEncrypt2Base ( blEncrypt );
	if ( ! chBase64 ) {
	    blEncrypt->Free ();
	    SER ( "! chBase64" );
	}
	blEncrypt->Free ();
	blEncrypt->Append ( chBase64 );
	free ( chBase64 );
    }//if ( base64 )

    return blEncrypt;
#else
    SER ( "no ssl build" );
#endif
}//_bnBlob *bnSSLEncrypt ( _bnBlob *Data, _bnBlob *Key, const char *password, int base64 )


_bnBlob *bnSSLDecrypt ( _bnBlob *Data, _bnBlob *Key, const char *password, int base64 ) {
    D9 ( "call" );
#ifdef  HAVE_SSL
    if ( ! Data      ) SER ( "! Data"      );
    if ( ! Data->len ) SER ( "! Data->len" );
    if ( ! Key       ) SER ( "! Key"       );
    if ( ! Key->len  ) SER ( "! Key->len"  );
    if (   password  ) SER ( "password set, bug2 src it out" );
    //Data->Show ( "Data" );//bug4debug

    OpenSSL_add_all_algorithms ();
    OpenSSL_add_all_ciphers    ();
    ERR_load_crypto_strings    ();

_bnBlob *blBase64 = NULL;
    if ( base64 ) {
	blBase64 = bnSSLBase2Encrypt ( (const char *)Data->data );
	if ( ! blBase64 ) SER ( "! blBase64" );
	//blBase64->Show ( "blBase64" );//bug4debug
	Data = blBase64;
    }//if ( base64 )

_bnBlob *blDecrypt = NULL;
char *TLSPrivateKey = (char *)Key->data;
    if      ( ( strlen ( TLSPrivateKey ) > strlen ( "-----BEGIN RSA PRIVATE KEY-----" ) ) && ( strncmp ( TLSPrivateKey, "-----BEGIN RSA PRIVATE KEY-----", strlen ( "-----BEGIN RSA PRIVATE KEY-----" ) ) == 0 ) ) {
	blDecrypt = bnSSLEncrypt2Sign ( Data, Key ); //decrypt with private -> client_rsa.key
    }
    else if ( ( strlen ( TLSPrivateKey ) > strlen ( "-----BEGIN PRIVATE KEY-----"     ) ) && ( strncmp ( TLSPrivateKey, "-----BEGIN PRIVATE KEY-----"    , strlen ( "-----BEGIN PRIVATE KEY-----"     ) ) == 0 ) ) {
	blDecrypt = bnSSLEncrypt2Sign ( Data, Key ); //decrypt with private -> client_rsa.key
    }
    else if ( ( strlen ( TLSPrivateKey ) > strlen ( "-----BEGIN CERTIFICATE-----"     ) ) && ( strncmp ( TLSPrivateKey, "-----BEGIN CERTIFICATE-----"    , strlen ( "-----BEGIN CERTIFICATE-----"     ) ) == 0 ) ) {
	blDecrypt = bnSSLSign2Data    ( Data, Key ); //decrypt with public -> my_rsa.crt
    }
    if ( blBase64 ) blBase64->Free ();
    if ( ! blDecrypt ) SER ( "! blDecrypt" );
    //blDecrypt->Show ( "blDecrypt" );//bug4debug

    return blDecrypt;
#else
    SER ( "no ssl build" );
#endif
}//_bnBlob *bnSSLDecrypt ( _bnBlob *Data, _bnBlob *Key, const char *password, int base64 )


char *bnSSLFingerprint ( const char *crtfile, const char *mode ) {
#ifdef  HAVE_SSL
    D9 ( "mode=~%s~ crtfile=~%s~", mode, crtfile );

_bnBlob Key;
    Key.FileRead ( crtfile );
    if ( ! Key.len ) SER ( "! Key.len" );
    //Key.ShowString ( "crtfile" );//bug4debug

_bnBlob *Cert = &Key;

BIO *bio = BIO_new ( BIO_s_mem () );
    BIO_write ( bio, Cert->data, Cert->len );
    if ( ! bio  ) SER ( "! bio"  );
X509 *x509 = PEM_read_bio_X509 ( bio, NULL, NULL, NULL );
    BIO_free ( bio );
    if ( ! x509 ) SER ( "! x509" );
    D9 ( "x509=%p", x509 );

char Fingerprint[1024];
    memset ( Fingerprint, 0x00, 1024 );

X509 *certificate = x509;

unsigned int n;
unsigned char md[EVP_MAX_MD_SIZE];
const EVP_MD *fdig = NULL;
    if      ( strcmp ( mode, "sha256" ) == 0 ) fdig = EVP_sha256 ();
    else if ( strcmp ( mode, "sha1"   ) == 0 ) fdig = EVP_sha1   ();
    else SER ( "unknown mode=~%s~", mode );
    if ( ! fdig ) SER ( "! fdig" );

    if ( ! X509_digest ( certificate, fdig, md, &n ) ) {
	X509_free ( certificate );
	SER ( "! X509_digest" );
    }
    X509_free ( certificate );
    D9 ( "n=%ld", n );

    for ( int j = 0; j < (int)n; j++ ) sprintf ( Fingerprint, "%s%02X:", Fingerprint, md[j] );
    if ( ! strlen ( Fingerprint ) ) SER ( "! strlen ( Fingerprint )" );
    Fingerprint[ strlen ( Fingerprint ) - 1 ] = 0;

    D9 ( "Fingerprint =~%s~", Fingerprint );
    return strdup ( Fingerprint );
#else
    SER ( "no ssl build" );
#endif

}//char *bnSSLFingerprint ( const char *crtfile, const char *mode )


//#define bnSSL_MAIN
#ifdef  bnSSL_MAIN

#include <bnRC4.h>


char Error[1024] = "";


char *MainMain ( int argc, char **argv ) {
    D9 ( "call" );

_bnRC4	bnRC;
	bnRC.Argc     = argc;
	bnRC.Argv     = argv;
	bnRC.ReadENV  = "certificate;password;default_days;default_bits;commonName;organizationalUnitName;organizationName;localityName;stateOrProvinceName;countryName;emailAddress;subjectAltName;subjectKeyIdentifier;authorityKeyIdentifier;basicConstraints;keyUsage;extendedKeyUsage;cacertificate;cakey;capassword";
	bnRC.Open (); SIER ( bnRC.Error );
#ifdef bnSSL_DEBUG
	bnRC.Show ();
#endif

    if ( bnRC.GetVal ( "encrypt"  ) ) {
    const char *inputfile  = bnRC.GetVal ( "encrypt"    );
    const char *outputfile = bnRC.GetVal ( "output"     );
    const char *keycrtfile = bnRC.GetVal ( "keycrtfile" );
    const char *password   = bnRC.GetVal ( "password"   );
    int base64 = 1;
	if ( ! inputfile  ) SER ( "! inputfile"  );
	if ( ! outputfile ) SER ( "! outputfile" );
	if ( ! keycrtfile ) SER ( "! keycrtfile" );
	D9 ( "inputfile=%s" , inputfile  );
	D9 ( "outputfile=%s", outputfile );
	D9 ( "keycrtfile=%s", keycrtfile );
    _bnBlob input;
	input.FileRead ( inputfile );
    _bnBlob Key;
	Key.FileRead ( keycrtfile );
	//Key.ShowString ( "keycrtfile" );//bug4debug
    _bnBlob *output = bnSSLEncrypt ( &input, &Key, password, base64 );
	if ( Error[0] ) return NULL;
	if ( ! output ) SER ( "! output" );
	output->FileWrite ( outputfile );
	output->Free ();
	return NULL;
    }//if ( bnRC.GetVal ( "encrypt" ) )

    if ( bnRC.GetVal ( "decrypt"  ) ) {
    const char *inputfile  = bnRC.GetVal ( "decrypt"    );
    const char *outputfile = bnRC.GetVal ( "output"     );
    const char *keycrtfile = bnRC.GetVal ( "keycrtfile" );
    const char *password   = bnRC.GetVal ( "password"   );
    int base64 = 1;
	if ( ! inputfile  ) SER ( "! inputfile"  );
	if ( ! outputfile ) SER ( "! outputfile" );
	if ( ! keycrtfile ) SER ( "! keycrtfile" );
	D9 ( "inputfile=%s" , inputfile  );
	D9 ( "outputfile=%s", outputfile );
	D9 ( "keycrtfile=%s", keycrtfile );
    _bnBlob input;
	input.FileRead ( inputfile );
    _bnBlob Key;
	Key.FileRead ( keycrtfile );
	//Key.ShowString ( "keycrtfile" );//bug4debug
    _bnBlob *output = bnSSLDecrypt ( &input, &Key, password, base64 );
	if ( Error[0] ) return NULL;
	if ( ! output ) SER ( "! output" );
	output->FileWrite ( outputfile );
	output->Free ();
	return NULL;
    }//if ( bnRC.GetVal ( "decrypt" ) )

char mBuffer[1024] = "";

char *certificate     = NULL;
    if ( bnRC.GetVal ( "certificate" ) ) certificate = strdup ( bnRC.GetVal ( "certificate" ) );
    if ( ! certificate ) SER ( "! certificate" );
    D1 ( "certificate=~%s~", certificate );

    strcat ( mBuffer, certificate );
    if ( strcmp ( ".crt", mBuffer + strlen ( mBuffer ) - 4 ) == 0 ) mBuffer[strlen(mBuffer)-4] = 0;
    if ( strcmp ( ".pem", mBuffer + strlen ( mBuffer ) - 4 ) == 0 ) mBuffer[strlen(mBuffer)-4] = 0;
    strcat ( mBuffer, ".key" );

char *rsafile      = NULL;
char *rsapassword  = NULL;
int   default_bits = 0;

    if ( bnRC.GetVal ( "default_bits" ) ) default_bits = atol   ( bnRC.GetVal ( "default_bits" ) );
    if ( bnRC.GetVal ( "rsafile"      ) ) rsafile      = strdup ( bnRC.GetVal ( "rsafile"      ) );
    else                                  rsafile      = strdup ( mBuffer );
    if ( bnRC.GetVal ( "password"     ) ) rsapassword  = strdup ( bnRC.GetVal ( "password"     ) );

char *crtfile       = NULL;
char *crtconfig     = NULL;
char *cacrtfile     = NULL;
char *carsafile     = NULL;
char *carsapassword = NULL;

                                            crtfile       = strdup ( certificate );
                                            crtconfig     = bnRC.cString ();
    if ( bnRC.GetVal ( "cacertificate"  ) ) cacrtfile     = strdup ( bnRC.GetVal ( "cacertificate" ) );
    if ( bnRC.GetVal ( "cakey"          ) ) carsafile     = strdup ( bnRC.GetVal ( "cakey"         ) );
    if ( bnRC.GetVal ( "capassword"     ) ) carsapassword = strdup ( bnRC.GetVal ( "capassword"    ) );

char *rsafileRX = bnSSLGenRSAKey ( rsafile, rsapassword, default_bits );
    if ( Error[0] ) return NULL;
    D2 ( "rsafileRX=~%s~", rsafileRX );

char *crtfileRX = bnSSLGenCRT ( crtfile, crtconfig, rsafile, rsapassword, cacrtfile, carsafile, carsapassword );
    if ( Error[0] ) return NULL;
    D2 ( "crtfileRX=~%s~", crtfileRX );

    return NULL;
}//char *MainMain ( int argc, char **argv )


int main ( int argc, char **argv ) {
    albus_log_level  = 2;
    albus_log_stdio  = strdup ( "stderr" );
#ifdef bnSSL_DEBUG
    albus_log_level  = 9;
#endif
    D9 ( "call" );
    MainMain ( argc, argv );
    if ( Error[0] ) {
	fprintf ( stderr, "Error: %s\n", Error );
	return -1;
    }
    return 0;
}//int main ( int argc, char **argv )
#endif

#endif
