//Copyright 2016-2018 blacknoise.at
//
//   Licensed under the Apache License, Version 2.0 (the "License");
//   you may not use this file except in compliance with the License.
//   You may obtain a copy of the License at
//
//       http://www.apache.org/licenses/LICENSE-2.0
//
//   Unless required by applicable law or agreed to in writing, software
//   distributed under the License is distributed on an "AS IS" BASIS,
//   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//   See the License for the specific language governing permissions and
//   limitations under the License.
//
//Version: 1.0.0
#ifdef __cplusplus
extern "C" {
#endif
#ifndef albus_log_H
#define albus_log_H
#include <stdlib.h>
#ifdef __APPLE__
# include <malloc/malloc.h>
#else
# include <malloc.h>
#endif


#define AL_ER 0x0F
#define AL_D1 0x01
#define AL_D2 0x02
#define AL_D3 0x03
#define AL_D4 0x04
#define AL_D5 0x05
#define AL_D6 0x06
#define AL_D7 0x07
#define AL_D8 0x08
#define AL_D9 0x09

extern int   albus_log_level;
extern int   albus_log_syslog;
extern char *albus_log_stdio;
extern char *albus_log_file;
extern char *albus_log_modul;
#define ModulNme albus_log_modul

#define ER(a01,...)  { albus_log ( AL_ER, Error, ModulNme, (char *)__PRETTY_FUNCTION__, __LINE__, a01, ## __VA_ARGS__ ); }
#define SEGT(a01,...){ albus_log ( AL_ER, Error, ModulNme, (char *)__PRETTY_FUNCTION__, __LINE__, a01, ## __VA_ARGS__ ); goto my_return; }
#define SER(a01,...) { albus_log ( AL_ER, Error, ModulNme, (char *)__PRETTY_FUNCTION__, __LINE__, a01, ## __VA_ARGS__ ); return NULL; }
#define PE(a01,...)  { albus_log ( AL_ER, Error, ModulNme, (char *)__PRETTY_FUNCTION__, __LINE__, a01, ## __VA_ARGS__ ); Error[0]=0;           }
#define PEC(a01,...) { albus_log ( AL_ER, Error, ModulNme, (char *)__PRETTY_FUNCTION__, __LINE__, a01, ## __VA_ARGS__ ); Error[0]=0; continue; }
#define EX(a01,...)  { albus_log ( AL_ER,  NULL, ModulNme, (char *)__PRETTY_FUNCTION__, __LINE__, a01, ## __VA_ARGS__ ); exit (1); }
#define D1(a01,...)  { albus_log ( AL_D1,  NULL, ModulNme, (char *)__PRETTY_FUNCTION__, __LINE__, a01, ## __VA_ARGS__ ); }
#define D2(a01,...)  { albus_log ( AL_D2,  NULL, ModulNme, (char *)__PRETTY_FUNCTION__, __LINE__, a01, ## __VA_ARGS__ ); }
#define D3(a01,...)  { albus_log ( AL_D3,  NULL, ModulNme, (char *)__PRETTY_FUNCTION__, __LINE__, a01, ## __VA_ARGS__ ); }
#define D4(a01,...)  { albus_log ( AL_D4,  NULL, ModulNme, (char *)__PRETTY_FUNCTION__, __LINE__, a01, ## __VA_ARGS__ ); }
#define D5(a01,...)  { albus_log ( AL_D5,  NULL, ModulNme, (char *)__PRETTY_FUNCTION__, __LINE__, a01, ## __VA_ARGS__ ); }
#define D6(a01,...)  { albus_log ( AL_D6,  NULL, ModulNme, (char *)__PRETTY_FUNCTION__, __LINE__, a01, ## __VA_ARGS__ ); }
#define D7(a01,...)  { albus_log ( AL_D7,  NULL, ModulNme, (char *)__PRETTY_FUNCTION__, __LINE__, a01, ## __VA_ARGS__ ); }
#define D8(a01,...)  { albus_log ( AL_D8,  NULL, ModulNme, (char *)__PRETTY_FUNCTION__, __LINE__, a01, ## __VA_ARGS__ ); }
#define D9(a01,...)  { albus_log ( AL_D9,  NULL, ModulNme, (char *)__PRETTY_FUNCTION__, __LINE__, a01, ## __VA_ARGS__ ); }

extern long albus_log_mi_last_free;
extern long albus_log_mi_last_offset;
#ifdef WIN32
//2do MI @ WIN32
#define MI(a01,...)  {}
#else
#define MI(a01,...)  { \
char a02[255] = ""; \
struct mallinfo MallInfo; \
    MallInfo = mallinfo (); \
    MallInfo.uordblks -= albus_log_mi_last_offset; \
    sprintf ( a02 ,"mi: used=%8d, free=%8d=%8d diff=%08ld", MallInfo.uordblks, MallInfo.fordblks, MallInfo.arena, albus_log_mi_last_free - MallInfo.uordblks ); \
    albus_log_mi_last_free = MallInfo.uordblks; \
    albus_log ( AL_D7,  NULL, ModulNme, (char *)__PRETTY_FUNCTION__, __LINE__, a02 ); \
}
#endif

#ifndef PFL
#ifndef CRLF
#	define CRLF "<br>\n"
#endif
#define PFL { fprintf ( stderr, "%70s[%04ld]%s", __PRETTY_FUNCTION__, (long)__LINE__, CRLF ); fflush ( stderr ); albus_log ( AL_D8,  NULL, ModulNme, NULL, 0, "%70s[%04ld]", __PRETTY_FUNCTION__, __LINE__ ); }
# ifdef DEBUG
#	define PFLID PFL
# else
#	define PFLID
# endif
#endif

extern void albus_log      ( int l_level, char *Error, char *Modul, const char *Function, long FunctionLine, const char *format, ... );
extern char *albus_log_set ( const char *Key, const char *Val );

#define D9TB(tb) if ( 9 == albus_log_level ) { tb->StdOut = stderr; tb->ShowHTML(); }
#define D8TB(tb) if ( 8 == albus_log_level ) { tb->StdOut = stderr; tb->ShowHTML(); }

#define PIE(sErr)   { if ( sErr[0] ) { ER(sErr); sErr[0]=0; } }
#define PIEC(sErr)  { if ( sErr[0] ) { ER(sErr); sErr[0]=0; continue; } }
#define PIER(sErr)  { if ( sErr[0] ) { ER(sErr); sErr[0]=0; return NULL; } }
#define	SET_ERROR(sErr) { sprintf ( Error, "%s @ %s[%ld]\n", sErr, __PRETTY_FUNCTION__, (long int)__LINE__ ); }
#define	SET_IF_ERROR(sErr) if ( sErr[0] ) { SET_ERROR(sErr); }
#define SET_IF_ERROR_RETURN(sErr) SET_IF_ERROR(sErr); if ( Error[0] ) { return NULL; }
#define SIER(sErr) SET_IF_ERROR_RETURN(sErr)
#define SIEPR(sErr)  { if ( sErr[0] ) { ER(sErr); return NULL; } }

#endif
#ifdef __cplusplus
}
#endif

//printf ( "%s"		, (char *)__PRETTY_FUNCTION__	); printf ( "	%s"	, (char *)__FILE__		);
