//Copyright 2016-2017 blacknoise.at
//
//   Licensed under the Apache License, Version 2.0 (the "License");
//   you may not use this file except in compliance with the License.
//   You may obtain a copy of the License at
//
//       http://www.apache.org/licenses/LICENSE-2.0
//
//   Unless required by applicable law or agreed to in writing, software
//   distributed under the License is distributed on an "AS IS" BASIS,
//   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//   See the License for the specific language governing permissions and
//   limitations under the License.
//
#ifndef bnBlob_H
#define bnBlob_H

#ifdef DEBUG
# define bnBlob_DEBUG
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef __APPLE__
# include <malloc/malloc.h>
#else
# include <malloc.h>
#endif
#include <stdarg.h>

#ifdef bnBlob_DEBUG
# include <albus_log.h>
#endif


void _bnBlobTest ( void );

class _bnBlob {
public:


    void FileRead ( const char *format, ... ) {

    va_list arg;
	va_start ( arg, format );
	vsprintf ( FileName, format, arg );
	va_end   ( arg );
#ifdef bnBlob_DEBUG
	D8 ( "this=%p FileName=~%s~", this, FileName );
#endif

_bnBlob srcBlob(1024);
    FILE *file_rx = NULL;
	if ( strcmp ( "stdin", FileName ) == 0 ) file_rx = stdin;
	else                                     file_rx = fopen ( FileName, "rb" );

	if ( ! file_rx ) return;
	while ( ( srcBlob.len = fread ( srcBlob.data, 1, 1024, file_rx ) ) ) {
	    Append ( &srcBlob );
	}
	fclose ( file_rx );

    }//void FileRead ( const char *format, ... )


    void FileWrite ( const char *format, ... ) {

    va_list arg;
	va_start ( arg, format );
	vsprintf ( FileName, format, arg );
	va_end   ( arg );
#ifdef bnBlob_DEBUG
	D8 ( "this=%p FileName=~%s~", this, FileName );
#endif
    FILE *file_tx_out = fopen ( FileName, "wb" );
	if ( ! file_tx_out ) return;
	fwrite ( data, 1, len, file_tx_out );
	fclose ( file_tx_out );

    }//void FileWrite ( const char *format, ... )


    void Clean ( void ) {
#ifdef bnBlob_DEBUG
	D8 ( "this=%p", this );
#endif
	len = 0;
	memset ( data + len, 0x00, size - len );
    }//void Clean ( void )


    void Append ( _bnBlob *srcBlob ) {
#ifdef bnBlob_DEBUG
	D8 ( "this=%p", this );
#endif
	Malloc ( srcBlob->len + 1 ); // +1 zero terminated string
	memcpy ( data + len, srcBlob->data, srcBlob->len );
	len += srcBlob->len;
    }//void Append ( _bnBlob *srcBlob )


    void Append ( const char *sAppend ) {
#ifdef bnBlob_DEBUG
	D8 ( "this=%p sAppend=%p", this, sAppend );
#endif
    unsigned long ReSize = strlen ( sAppend ) + 1;
	Malloc ( ReSize );
	sprintf ( (char *)(data + len), "%s%s", (char *)(data + len), sAppend );
	len += strlen ( sAppend );
    }//void Append ( char *sAppend )


    void Free ( void ) {
#ifdef bnBlob_DEBUG
	D8 ( "this=%p", this );
#endif
	if ( data ) free ( data );
	memset ( this, 0x00, sizeof ( _bnBlob ) );
    }//void Clean ( void )


    void Malloc ( unsigned long ReMalloc ) {
	if ( ! ReMalloc ) ReMalloc = 0xFF;
	if ( ( size - len ) >= ReMalloc ) return;
#ifdef bnBlob_DEBUG
	D8 ( "this=%p ReMalloc=%06ld", this, ReMalloc );
#endif
	size += ReMalloc;
	data  = (unsigned char *)realloc ( data, size );
	memset ( data + len, 0x00, size - len );
    }//void Malloc ( unsigned long ReMalloc )


    _bnBlob ( const char *sAppend ) {
#ifdef bnBlob_DEBUG
	D8 ( "this=%p sAppend=%p", this, sAppend );
#endif
	memset ( this, 0x00, sizeof ( _bnBlob ) );
	Append ( sAppend );
    }//_bnBlob ( const char *sAppend )

    _bnBlob ( _bnBlob *srcBlob ) {
#ifdef bnBlob_DEBUG
	D8 ( "this=%p", this );
#endif
	memset ( this, 0x00, sizeof ( _bnBlob ) );
	Append ( srcBlob );
    }//_bnBlob ( const char *sAppend )


    _bnBlob ( unsigned long ReMalloc ) {
#ifdef bnBlob_DEBUG
	D8 ( "this=%p ReMalloc=%06ld", this, ReMalloc );
#endif
	memset ( this, 0x00, sizeof ( _bnBlob ) );
//malloc = ReMalloc;
	Malloc ( ReMalloc );
    }//_bnBlob ( unsigned long ReMalloc )


    void Show ( const char *format, ... ) {
    char sString[1024];
    va_list arg;
	va_start ( arg, format );
	vsprintf ( sString, format, arg );
	va_end   ( arg );
#ifdef bnBlob_DEBUG
	D8 ( "this=%p sString=~%s~", this, sString );
#endif
	fprintf ( stderr, "_bnBlob%20s this=%p len=%06ld data=%p size=%06ld\n"  , sString, this, len, data, size ); fflush ( stderr );
    }//void Show ( const char *format, ... )


    void ShowString ( const char *format, ... ) {
    char sString[1024];
    va_list arg;
	va_start ( arg, format );
	vsprintf ( sString, format, arg );
	va_end   ( arg );
#ifdef bnBlob_DEBUG
	D8 ( "this=%p sString=~%s~", this, sString );
#endif
	fprintf ( stderr, "_bnBlob%20s this=%p len=%06ld data=~%s~ size=%06ld\n", sString, this, len, data, size ); fflush ( stderr );
    }//void ShowString ( const char *format, ... )


    _bnBlob() {
#ifdef bnBlob_DEBUG
	D8 ( "this=%p", this );
#endif
	memset ( this, 0x00, sizeof ( _bnBlob ) );
    }


    ~_bnBlob() {
#ifdef bnBlob_DEBUG
	D8 ( "this=%p", this );
#endif
	Free ();
    }

    unsigned long  size;
    unsigned long  len;
    unsigned char *data;
    char FileName[2048];

};

#endif
