//Copyright 2016-2018 blacknoise.at
//
//   Licensed under the Apache License, Version 2.0 (the "License");
//   you may not use this file except in compliance with the License.
//   You may obtain a copy of the License at
//
//       http://www.apache.org/licenses/LICENSE-2.0
//
//   Unless required by applicable law or agreed to in writing, software
//   distributed under the License is distributed on an "AS IS" BASIS,
//   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//   See the License for the specific language governing permissions and
//   limitations under the License.
//
#ifndef bnForward_H
#define bnForward_H

#ifdef DEBUG
#define bnForward_DEBUG
#endif

#include <string>

#include <bnSocket.h>
#ifdef bnForward_DEBUG
# include <albus_log.h>
#endif

#ifndef bn_delete
#define bn_delete(a) { if (a) { delete a; a = NULL; } }
#endif


class _bnForwardMeta {//2d mv -> class _Conn
public:

    char                Error[1024];
_bnBlob                *LocalRX;
_bnBlob                *LocalTX;
_bnBlob                *RemoteRX;
_bnBlob                *RemoteTX;
_bnSocket              *soLocal;
_bnSocket              *soRemote;
    long                Status;
    time_t              Timeout;
    char               *CONNECT;

    _bnForwardMeta () {
#ifdef bnForward_DEBUG
	D8 ( "this=%p", this );
#endif
	memset ( this     , 0x00, sizeof ( _bnForwardMeta ) );
    }

    ~_bnForwardMeta () {
#ifdef bnForward_DEBUG
	D8 ( "this=%p", this );
	bn_delete ( LocalRX  );
	bn_delete ( LocalTX  );
	bn_delete ( RemoteRX );
	bn_delete ( RemoteTX );
	bn_delete ( soRemote );
	bn_delete ( CONNECT  );
#endif
    }

};


class _bnForward {
public:
    char *Set     ( long *Idx, const std::string &Key, long *KeyIdx, const std::string &Val );
    char *Loop    ( void );
    char *Close   ( void );
    void *Connect ( void );

    char                Error[1024];
    char                InitDone;
//Set Vars
    char               *RemoteHost;
    unsigned short      RemotePort;
    unsigned short      RemoteTLS;
    char               *RemoteTLSCertificate;
    char               *RemoteTLSPrivateKey;
    char               *RemoteTLSPrivateKeyPass;
    char               *RemoteTLSFingerprint00;
    char               *RemoteTLSFingerprint01;
    char               *RemoteTLSFingerprint02;
    char               *RemoteTLSFingerprint03;
    char               *RemoteTLSFingerprint04;
    unsigned short      LocalListen;
    char               *LocalHost;
    unsigned short      LocalPort;
    unsigned short      LocalTLS;
    char               *LocalTLSCertificate;
    char               *LocalTLSPrivateKey;
    char               *LocalTLSPrivateKeyPass;
    unsigned short      ProxyConnection;
    char               *UserAgent;
    char               *ConnectRequest;
    unsigned short      PrintPID;
    unsigned short      PrintPort;
    unsigned short      CloseAndExit;

    _bnForward () {
#ifdef bnForward_DEBUG
	D8 ( "this=%p", this );
#endif
	memset ( this     , 0x00, sizeof ( _bnForward ) );
	soServer          = new _bnSocket;
	LocalListen       =    200;
	uSleepMin         =   1000;
	uSleepMax         = 200000;
    }

    ~_bnForward () {
#ifdef bnForward_DEBUG
	D8 ( "this=%p", this );
#endif
	Close ();
	bn_delete ( RemoteHost              );
	bn_delete ( RemoteTLSCertificate    );
	bn_delete ( RemoteTLSPrivateKey     );
	bn_delete ( RemoteTLSPrivateKeyPass );
	bn_delete ( RemoteTLSFingerprint00  );
	bn_delete ( RemoteTLSFingerprint01  );
	bn_delete ( RemoteTLSFingerprint02  );
	bn_delete ( RemoteTLSFingerprint03  );
	bn_delete ( RemoteTLSFingerprint04  );
	bn_delete ( LocalHost               );
	bn_delete ( LocalTLSCertificate     );
	bn_delete ( LocalTLSPrivateKey      );
	bn_delete ( LocalTLSPrivateKeyPass  );
	bn_delete ( UserAgent               );
	bn_delete ( ConnectRequest          );
	delete soServer;
    }

private:
    _bnSocket          *soServer;
    unsigned long       uSleep;
    unsigned long       uSleepMin;
    unsigned long       uSleepMax;

};

#endif
