//Copyright 2016-2018 blacknoise.at
//
//   Licensed under the Apache License, Version 2.0 (the "License");
//   you may not use this file except in compliance with the License.
//   You may obtain a copy of the License at
//
//       http://www.apache.org/licenses/LICENSE-2.0
//
//   Unless required by applicable law or agreed to in writing, software
//   distributed under the License is distributed on an "AS IS" BASIS,
//   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//   See the License for the specific language governing permissions and
//   limitations under the License.
//
#ifndef bnSocket_H
#define bnSocket_H

#ifdef DEBUG
# define bnSocket_DEBUG
#endif

#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <vector>

//2do rm #ifdef ANDROID
//2do rm # include <linux/in.h>
//2do rm # include <sys/endian.h>
//2do rm #endif
#ifndef WIN32
//2do rm # include <sys/socket.h>
//2do rm # include <sys/select.h>
# include <arpa/inet.h>
# include <netinet/in.h>
# include <netdb.h>
#else
# include <winsock2.h>
#endif

#ifdef HAVE_SSL
# include <openssl/ssl.h>
# include <openssl/err.h>
#else
# define SSL     void
# define SSL_CTX void
#endif

#include <bnBlob.h>

#ifdef bnSocket_DEBUG
# include <albus_log.h>
#endif

#ifndef bn_delete
#define bn_delete(a) { if (a) { delete a; a = NULL; } }
#endif



class _bnSocket {
public:

    void *(*CloseCB)( _bnSocket *, void * );
    void   *CloseCBArg;
    void *(*AcceptCB)( _bnSocket *, void * );
    void   *AcceptCBArg;

    char *Open  ( void      );
    char *Loop  ( void      );
    void *Close ( void      );
    void *Accept( void *    );
    void *Read  ( void *    );
    void *Write ( void *    );
    void *Save  ( char *Mode, long Slot, _bnBlob *rx );
    void *FingerprintCheck ( SSL *ssl );

    int                   fdSocket;
    _bnBlob              *tx;
    _bnBlob              *txThis;
    unsigned long         txLen;
    _bnBlob              *rx;
    _bnBlob              *rxThis;
    unsigned long         rxLen;
    char                  fdOpen;
    char                  fdClose;
    SSL                  *ssl;
    unsigned long         Event;

    char                  Error[1024];
    char                 *Host;
    unsigned short        Port;
    unsigned short        Listen;
    struct hostent       *hostbyname;
    struct sockaddr_in   *Sockaddr;
std::vector<_bnSocket *> *vs_stack;
    void                 *Meta;
    fd_set                the_state; //2do check memset ( this ...
    int                   fd_last;
    long                 *SaveLoop;
    char                 *SaveFile;
    SSL_CTX              *ssl_ctx;
    char                  TLS;
    char                 *TLSCertificate;    //Local
    char                 *TLSPrivateKey;     //Local
    char                 *TLSPrivateKeyPass; //Local
    char                 *TLSCACertificate;  //Remote Certs
    char                 *TLSFingerprint00;  //Remote
    char                 *TLSFingerprint01;  //Remote
    char                 *TLSFingerprint02;  //Remote
    char                 *TLSFingerprint03;  //Remote
    char                 *TLSFingerprint04;  //Remote
    char                  AutoOpen;
    //long  Retry;
    //long  RetryNext;
    //long  *Now;

    _bnSocket () {
#ifdef bnSocket_DEBUG
	D8 ( "this=%p", this );
#endif
	memset (      this, 0x00, sizeof ( _bnSocket ) );
	vs_stack = new std::vector<_bnSocket *>;
	//Retry    =   30;
    }

    ~_bnSocket () {
#ifdef bnSocket_DEBUG
	D8 ( "this=%p", this );
#endif

	fdClose = 1;
	Close ();
	if ( rx == rxThis ) { bn_delete ( rxThis ); rx = NULL; }
	if ( tx == txThis ) { bn_delete ( txThis ); tx = NULL; }
	bn_delete ( Host              );
	bn_delete ( SaveFile          );
	bn_delete ( TLSCertificate    );
	bn_delete ( TLSPrivateKey     );
	bn_delete ( TLSPrivateKeyPass );
	bn_delete ( TLSCACertificate  );
	bn_delete ( TLSFingerprint00  );
	bn_delete ( TLSFingerprint01  );
	bn_delete ( TLSFingerprint02  );
	bn_delete ( TLSFingerprint03  );
	bn_delete ( TLSFingerprint04  );
	//2do free hostbyname
	bn_delete ( Sockaddr          );

#ifdef HAVE_SSL
	if ( ssl     ) { SSL_free     ( ssl     ); ssl     = NULL; }
	if ( ssl_ctx ) { SSL_CTX_free ( ssl_ctx ); ssl_ctx = NULL; }
#endif
	delete vs_stack;
    }//~_bnSocket ()

private:
    char *OpenClient ( void );
    char *OpenServer ( void );

};

#endif
